import numpy as np
import pandas as pd
import os
import emcee

from emcee import EnsembleSampler
from .func import *
from .classes import *
from numba import njit
from scipy.interpolate import interp1d

#########################################################
#														#
#		PHOTOMETRIC + SPEC VERSION OF THE FITTING		#
#														#
#########################################################
@njit
def lnpostfn_spec_noAGN(theta, P, modelBG, modelSG, fluxFit, efluxFit, UL, wei, z, wavFit):
	
	"""
    This function calculates the log-likelihood between spectral + photometric data and model without AGN contribution.
    ------------
    :param theta: vector containing the parameters.
    :param P: vector containing the priors for each of the parameters.
    :param modelBG: model dust continuum template.
    :param modelSG: model PAH template.
    :param fluxFit: observed fluxes.
    :param efluxFit: uncertainties on the observed fluxes.
    :param UL: vector contaning the upper-limits, if any.
    :param wei: weighting of the data points.
    :param z: redshift.
	:param wavFit: observed wavelengths.
    ------------
    :return logl: log-likelihood of the model knowing the parameters theta.
    """

    # Set a variable for the loglikelihood
	logl = 0

	# Test if the sum of the wei is equal to 1, which is used as a flag for the final calculation of the loglikelihood of the model.
	# If it is the final calculation (i.e. paraneters have been optimised) we do not account for the priors.	
	if wei[0] != 1.:
		# Prior constraint on the dust normalisation
		logl += -0.5*(theta[0])**2./P[0][1]**2.

		# Prior constraint on the PAH emission
		logl += -0.5*(theta[1])**2./P[1][1]**2.
		# Constraint on the 0.3 dex between the PAH and the dust continuum
		logl += (-0.5*(theta[1] + P[1][0] - (0.24 + 0.86 * (theta[0] + P[0][0])))**2./0.1**2.)*50.
	
	# Define the model knowing the parameters theta
	ymodel = 10**(theta[0] + P[0][0]) * modelBG + 10**(theta[1] + P[1][0]) * modelSG

	# Calculate the log-likelihood for the upper limits
	o = np.where(UL == 1)[0]
	if len(o) > 0:
		logl += loglUL(ymodel[o], fluxFit[o])

	# Calculate the log-likelihood for the detected values
	o = np.where(UL == 0)[0]
	logl += logldet(ymodel[o], fluxFit[o], efluxFit[o], wei[o])

	return logl

@njit
def lnpostfn_spec_wAGN(theta, P, modelBG, modelSG, fluxFit, efluxFit, UL, wei, z, wavFit, modelAGN):

	"""
    This function calculates the log-likelihood between spectral + photometric data and model that includes AGN contribution.
    ------------
    :param theta: vector containing the parameters.
    :param P: vector containing the priors for each of the parameters.
    :param modelBG: model dust continuum template.
    :param modelSG: model PAH template.
    :param fluxFit: observed fluxes.
    :param efluxFit: uncertainties on the observed fluxes.
    :param UL: vector contaning the upper-limits, if any.
    :param wei: weighting of the data points.
    :param z: redshift.
    :param wavFit: observed wavelengths.
    :param modelAGN: AGN model using the priors from the fit to calculate the final log-likelihood.
    ------------
    :return logl: log-likelihood of the model knowing the parameters theta.
    """
	
	# set a variable for the log-likelihood
	logl = 0

	# Test if the sum of the wei is equal to 1, which is used as a flag for the final calculation of the loglikelihood of the model.
	# If it is the final calculation (i.e. paraneters have been optimised) we do not account for the priors.	
	if wei[0] != 1:
		# Normal prior on the norm of the dust
		logl += -0.5*(theta[0])**2./P[0][1]**2.

		# Normal prior on the norm of the PAH
		logl += -0.5*(theta[1])**2./P[1][1]**2.
		# Constraint on the 0.3 dex between the dust contrinuum and the PAH emission
		logl += (-0.5*(theta[1] + P[1][0] - (0.24 + 0.86 * (theta[0] + P[0][0])))**2./0.1**2.)*50.

		# Normal prior on the AGN power law
		logl += -0.5*(theta[2])**2./P[2][1]**2.

		# Normal prior on alpha1 + boundaries (i.e. -4.<alpha1<4.)
		if (theta[3]+P[3][0] < -1) | (theta[3]+P[3][0] > 4.) == True:
				return -np.inf
		logl += -0.5*(theta[3])**2./P[3][1]**2.

		# Normal prior on alpha2 + boundaries (i.e. -4.<alpha2<2.)
		if (theta[4]+P[4][0] < -4.) | (theta[4]+P[4][0] > 2.) == True:
				return -np.inf
		logl += -0.5*(theta[4])**2./P[4][1]**2.

		# Normal Prior on the normalisation of the Si emission at 11 microns
		logl += -0.5*(theta[5])**2./P[5][1]**2.

		# Normal Prior on the normalisation of the Si emission at 18 microns
		logl += -0.5*(theta[6])**2./P[6][1]**2.

		# Prior on the position of the break + boundaries (i.e. 15<lbreak<100.)
		# Prior is flat below the best guess for lbreak and fall at wavelengths beyond that value.
		if ((theta[7]+P[7][0]) <= 15.) | (theta[7]+P[7][0] >= 100.) == True:
				return -np.inf
		sigma = 0.434 * P[7][1]/P[7][0]
		x = (np.log10(np.array(theta[7]+P[7][0])) - np.log10(P[7][0]))/np.sqrt(2.)/sigma
		logl += np.log(1. - 0.5 * (1. + erf_approx(np.array([x]))))

		# Calculate the model for the AGN given parameters theta.
		modelPL = 10**(theta[2] + P[2][0]) * AGNmodel_jit(wavFit, 15.*(1.+z), theta[7]+P[7][0]*(1.+z), theta[3]+P[3][0], theta[4]+P[4][0], -3.5)
		modelSi11 = 10**(theta[5] + P[5][0]) * Gauss_jit(wavFit, np.log10(11.*(1.+z)), 0.05)
		modelSi18 = 10**(theta[6] + P[6][0]) * Gauss_jit(wavFit, np.log10(18.*(1.+z)), 0.07)

		# Calculate the full model (galaxy + AGN) given parameters theta.
		ymodel = modelPL + modelSi11 + modelSi18 + 10**(theta[0] + P[0][0]) * modelBG + 10**(theta[1]+ P[1][0]) * modelSG
	else:
		# if final calculation of the log-likelihood (i.e. parameters have been optimised), the models (galaxy and AGN) have been
		# calculated prior to this (i.e. kind of like templates, since the IRS has been reduced to at least 3 data points) and we just combine them in a full model.
		ymodel = 10**(theta[0] + P[0][0]) * modelBG + 10**(theta[1]+ P[1][0]) * modelSG + modelAGN

	# Calculate the log-likelihood for the undetected values
	o = np.where(UL == 1)[0]
	if len(o) > 0:
		logl += loglUL(ymodel[o], fluxFit[o])

	# Calculate the log-likelihood for detected values
	o = np.where(UL == 0)[0]
	logl += logldet(ymodel[o], fluxFit[o], efluxFit[o], wei[o])
	
	return logl

def runSEDspecFit(wavSpec, fluxSpec, efluxSpec,\
				  wavPhot, fluxPhot, efluxPhot, filters, \
				  z = -0.01,\
				  ULPhot = [], \
				  obsCorr = True,\
				  S9p7_fixed = -99., \
				  Nmc = 10000, pgrbar = 1, \
				  Pdust = [10., 3.], PPAH = [9., 3.], Ppl = [-1., 3.], Pbreak = [40., 1.], Pslope1 = [1., 2.], Pslope2 = [1., 2.], \
				  Plsg = [-1., 3.], Pllg = [-1., 3.],\
				  templ = ''):

	"""
    This function fits the observed SED when a spectrum is combined to photometric data. The observed wavelengths, fluxes and uncertainties on the fluxes are passed separately for the spectrum and the photometric data.
    ------------
    :param wavSpec: observed wavelengths for the spectrum (in microns).
    :param fluxSpec: observed fluxes for the spectrum (in Jansky).
    :param efluxSpec: observed uncertainties on the fluxes for the spectrum (in Jansky).
    :param wavPhot: observed wavelengths for the photometry (in microns).
    :param fluxPhot: observed fluxes for the photometry (in Jansky).
    :param efluxPhot: observed uncertainties on the fluxes for the photometry (in Jansky).
    :param filters: name of the photometric filters to include in the fit.
    ------------
    :keyword z: redshift of the source. Default = 0.01.
    :keyword ULPhot: vector of length Nphot, where Nphot is the number of photometric data. If any of the value is set to 1, the corresponding flux is set has an upper limit in the fit. Default = [].
    :keyword obsCorr: if set to True, iragnsep attempt to calculate the total silicate absorption at 9.7micron, and to correct the observed fluxes for obscuration. Default = True
    :keyword S9p7_fixed: can be used to pass a fixed value for the total silicate absorption at 9.7 micron. Default = -99.
    :keyword Nmc: numer of MCMC run. Default = 10000.
    :keyword pgrbar: if set to 1, display a progress bar while fitting the SED. Default = 1.
    :keyword Pdust: normal prior on the log-normalisation of the galaxy dust continuum template. Default = [10., 3.] ([mean, std dev]).
    :keyword PPAH: normal prior on the log-normalisation of the PAH template. Default = [9., 3.] ([mean, std dev]).
	:keyword Ppl: normal prior on the log-normalisation AGN continuum (defined at 10 micron). Default = [-1., 3.] ([mean, std dev]).
	:keyword Pbreak: prior on lbreak, the position of the break. Default = [40., 1.] ([mean, std dev]).
	:keyword Pslope1: normal prior on alpha1, the slope of the first power-law defined between 1<lambda<15 microns. Default = [1., 2.] ([mean, std dev]).
	:keyword Pslope2: normal prior on alpha2, the slope of the second power-law defined between 15<lambda<lbreak. Default = [1., 2.] ([mean, std dev]).
	:keyword Plsg: normal prior on the log-normalisation of the silicate emission at 11micron. Default = [-1., 3.] ([mean, std dev]).
	:keyword Pllg: normal prior on the log-normalisation of the silicate emission at 18micron. Default = [-1., 3.] ([mean, std dev]).
	------------
    :return res_fit: dataframe containing the results of all the possible fits.
    :return res_fitBM: dataframe containing the results of the best fit only.
    """

	# If no template are used, open the Bernhard20 templates.
	if len(templ) == 0:
		path = os.path.dirname(iragnsep.__file__)
		templ = pd.read_csv(path+'/iragnsep_templ.csv')

	# Extract the names of the templates
	keys = templ.keys().values
	nameTempl_gal = []
	nameTempl_PAH = []
	for key in keys:
		if str(key).startswith('gal'):
			if str(key).endswith('PAH') == False:
				nameTempl_gal.append(key)
			else:
				nameTempl_PAH.append(key)

	# Test that we have template for everything (if no galaxy then it crashes)
	if len(nameTempl_gal) == 0:
		raise ValueError('The galaxy template does not exist. The name of the column defining nuLnu for the galaxy template needs to start with "gal".')

	# define the wavelengths
	try:
		wavTempl = templ['lambda_mic']
	except:
		raise ValueError('Rename the wavelengths column of the template "lambda_mic".')

	# Define the rest wavelengths for the photometric and the spectral data
	wavSpec_rest = wavSpec/(1. + z)
	wavPhot_rest = wavPhot/(1. + z)

	# Correct for absorption, using the total 9.7micron absorption feature.
	if S9p7_fixed == -99.:
		# Test if there are actually data in the wavelength range, if not, continue the fit without correcting for obscuration
		o = np.where((wavSpec_rest > 9.) & (wavSpec_rest < 10.))[0]
		if (len(o) == 0.) & (obsCorr == True):
			print('*******************')
			print('It has failed to correct for obscuration. The most likely explanation is redshift since it needs the restframe anchor ' + \
					   'wavelengths to measure the strength of the silicate absorption. The fit is continued without correcting for obscuration.')
			print('*******************')
			obsCorr = False

		# Correct for obscuration. If it somehow fails, continue the fit without correcting for obscuration.
		if obsCorr == True:
			try:
				_tau9p7 = obsC(wavSpec_rest, fluxSpec, efluxSpec)
				_tau9p7 = round(_tau9p7[0],3)
			except:
				_tau9p7 = -99.
				print('*******************')
				print('It has failed to correct the IRS spectrum for obscuration. The most likely explanation is redshift since it needs the restframe anchor ' + \
					   'wavelengths to measure the strength of the silicate absorption. The fit is continued without correcting for obscuration.')
				print('*******************')
				pass
		else:
			_tau9p7 = -99.
	else:
		if S9p7_fixed == -98.:
			_tau9p7 = -99.
		else:
			# If fixed, use tau9p7 fixed by the user
			tauvec = 10**np.arange(-5., 5., 0.01)
			S9p7vec = (1. - np.exp(-tauvec))/tauvec
			f = interp1d(S9p7vec, tauvec)

			S9p7 = np.exp(-S9p7_fixed)

			_tau9p7 = np.array([f(S9p7)])[0]

	if _tau9p7 > 0:
		tauSpec = KVTextinctCurve(wavSpec_rest) * _tau9p7
		tauPhot = KVTextinctCurve(wavPhot_rest) * _tau9p7

		fluxSpec_corr = fluxSpec/((1. - np.exp(-tauSpec))/tauSpec)
		efluxSpec_corr = efluxSpec/((1. - np.exp(-tauSpec))/tauSpec)
		fluxPhot_corr = fluxPhot/((1. - np.exp(-tauPhot))/tauPhot)
		efluxPhot_corr = efluxPhot/((1. - np.exp(-tauPhot))/tauPhot)
	else:
		fluxSpec_corr = fluxSpec
		efluxSpec_corr = efluxSpec
		fluxPhot_corr = fluxPhot
		efluxPhot_corr = efluxPhot
		_tau9p7 = -99.

	# Concatenate the spectral and the photometric data.
	wavFit = np.concatenate([wavSpec, wavPhot])
	fluxFit = np.concatenate([fluxSpec_corr, fluxPhot_corr])
	efluxFit = np.concatenate([efluxSpec_corr, efluxPhot_corr])

	#Calculate the weights that should be allocated to the spectra, with respect to the photometry
	Npoints = float(len(fluxFit))
	Photwei = 1./((Npoints-len(wavSpec))/Npoints)
	Specwei = 1./((len(wavSpec))/Npoints)
	Specwei_dist = np.zeros(len(wavSpec)) + Specwei
	Photwei_dist = np.zeros(len(wavPhot)) + Photwei
	wei = np.concatenate([Specwei_dist, Photwei_dist])/np.sum(np.concatenate([Specwei_dist, Photwei_dist]))*100.

	# Prepare the upper limits. Set a vector of zeros for the spectra. Set a vector fo zeros for the photometry if UL is underfined.
	if len(ULPhot) != len(wavPhot):
		ULPhot = np.zeros(len(wavPhot))
	ULSpec = np.zeros(len(wavSpec))
	UL = np.concatenate([ULSpec, ULPhot])

	# Calculate the number of points that the spectra should be reduced to for the final calculation of the log-likelihood of the model,
	# knowing the optimised parameters.
	photoCov = max(wavPhot) - max(wavSpec)
	photo_covPerDat = float(len(fluxPhot))/photoCov
	specCov = max(wavSpec) - min(wavSpec)
	Npoints_red = (photo_covPerDat * specCov) + 1
	if Npoints_red < 4:
		Npoints_red = 4

	wavSpecAnchor = 10**np.linspace(np.log10(min(wavSpec)), np.log10(max(wavSpec)), Npoints_red)
	wavSpec_red = []
	fluxSpec_red = []
	for i in range(0, Npoints_red-1):
		o = np.where((wavSpec >= wavSpecAnchor[i]) & (wavSpec < wavSpecAnchor[i+1]))[0]
		wavSpec_red.append(np.mean(wavSpec[o]))
		fluxSpec_red.append(np.mean(fluxSpec_corr[o]))
	efluxSpec_red = np.array(fluxSpec_red) * np.mean(efluxPhot/fluxPhot)

	# These variables contain the reduced observed wavelengths, fluxex, and uncertainties.
	wavFit_red = np.concatenate([wavSpec_red, wavPhot])
	fluxFit_red = np.concatenate([fluxSpec_red, fluxPhot_corr])
	efluxFit_red = np.concatenate([efluxSpec_red, efluxPhot_corr])
	ULSpec_red = np.zeros(len(wavSpec_red))
	UL_red = np.concatenate([ULSpec_red, ULPhot])
	wei_red = np.ones(len(wavFit_red))

	# Define the free parameters
	lsg_perTempl = [] #silicate at 11 micron
	elsg_perTempl = []
	llg_perTempl = [] #silicate at 18 micron
	ellg_perTempl = []
	lnAGN_perTempl = [] #norm AGN continuum
	elnAGN_perTempl = []
	coff_perTempl = [] #cut off or position of the break
	ecoff_perTempl = []
	alpha1_perTempl = [] #slope of the first power law
	ealpha1_perTempl = []
	alpha2_perTempl = [] #slope of the second power law
	ealpha2_perTempl = []
	
	lnDust_perTempl = [] #dust continuum
	elnDust_perTempl = []
	lnPAH_perTempl = [] #PAH emission
	elnPAH_perTempl = []
	
	logl_perTempl = [] #log-likelihood of the mode
	tplName_perTempl = []#Name of the template 
	tau9p7_save = []#total absorption at 9.7 micron
	AGNon = []#flag for the use of the AGN in the fit
	nParms = []# number of parameters

	# Fit looping over every of our galaxy templates
	for name_i in nameTempl_gal:
		assert isinstance(name_i, str), "The list nameTempl requests strings as it corresponds to the names" + \
										" given to the various templates of galaxies to use for the fit."

		if pgrbar == 1:
			print("****************************************")
			print("  Fit of "+name_i+" as galaxy template  ")
			print("****************************************")

		# Define synthetic fluxes for the dust continuum model at the observed wavelength to match the observed fluxes
		nuLnuBGTempl = templ[name_i].values

		Fnu = nuLnuToFnu(wavTempl, nuLnuBGTempl, z)
		fluxSpec_model = np.interp(wavSpec_rest, wavTempl, Fnu)

		SEDgen = modelToSED(wavTempl, nuLnuBGTempl, z)
		fluxPhot_model = []
		for filt in filters:
		 	fluxPhot_model.append(getattr(SEDgen, filt)())
		modelBG_phot = fluxPhot_model

		modelBG = np.concatenate([fluxSpec_model, fluxPhot_model])

		# Define synthetic fluxes for the PAH model at the observed wavelength to match the observed fluxes.
		# When an empirical template is used, define a vector of zeros so that no PAH emission is accounted for.
		if name_i.endswith('dust') == False:
			modelSG = np.zeros(len(modelBG))
			modelSG_phot = np.zeros(len(modelBG_phot))
		else:
			nuLnuSGTempl = templ[nameTempl_PAH[0]].values

			Fnu = nuLnuToFnu(wavTempl, nuLnuSGTempl, z)
			fluxSpec_model = np.interp(wavSpec_rest, wavTempl, Fnu)

			SEDgen = modelToSED(wavTempl, nuLnuSGTempl, z)
			fluxPhot_model = []
			for filt in filters:
				fluxPhot_model.append(getattr(SEDgen, filt)())
			modelSG_phot = fluxPhot_model

			modelSG = np.concatenate([fluxSpec_model, fluxPhot_model])

		# fit without the AGN
		ndim = 2 #Number of paraneter
		nwalkers = int(2. * ndim) #number of walkers

		# Define the starting point of each of the parameters. Flat distrib. between -1 and 1. Parameters are normalised with a zero mean to ease convergence.
		parms = np.zeros(shape=(nwalkers, ndim))
		parms[:,0] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm Dust
		parms[:,1] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm PAH

		# Set the ensemble sampler of Goodman&Weare and run the MCMC for Nmc steps.
		sampler = EnsembleSampler(nwalkers, ndim, lnpostfn_spec_noAGN, args = (np.array([Pdust, PPAH]), modelBG, modelSG, fluxFit, efluxFit, \
								  UL, wei, z, wavFit))
		sampler.run_mcmc(parms, Nmc, progress=bool(pgrbar))
		
		# Build the flat chain after burning 20% of it and thinning to every 10 values.
		NburnIn = int(0.2 * Nmc)
		chain = sampler.get_chain(discard=NburnIn, thin=10, flat=True)

		# Save the optimised parameters. Median of the posterior as best fitting value and std dev as 1sigma uncertainties.
		# Dust continuum
		lnDust_perTempl.append(round(np.median(chain[:,0]),3))
		elnDust_perTempl.append(round(np.std(chain[:,0]),3))

		# PAH continuum (-20. for empirical template since no PAH should be added).
		if np.sum(modelSG) > 0.:
			lnPAH_perTempl.append(round(np.median(chain[:,1]),3))
			elnPAH_perTempl.append(round(np.std(chain[:,1]),3))
		else:
			lnPAH_perTempl.append(-20.)
			elnPAH_perTempl.append(0.0)

		# Norm AGN, here -99. since no AGN accounted for
		lnAGN_perTempl.append(-99.)
		elnAGN_perTempl.append(-99.)

		# slope of the first power law, here -99. since no AGN accounted for
		alpha1_perTempl.append(-99.)
		ealpha1_perTempl.append(-99.)

		# slope of the second power law, here -99. since no AGN accounted for
		alpha2_perTempl.append(-99.)
		ealpha2_perTempl.append(-99.)

		# cut off or position of the break, here -99. since no AGN accounted for
		coff_perTempl.append(-99.)
		ecoff_perTempl.append(-99.)

		# silicate emisison at 11 microns, here -99. since no AGN accounted for
		lsg_perTempl.append(-99.)
		elsg_perTempl.append(-99.)

		# silicate emisison at 18 microns, here -99. since no AGN accounted for
		llg_perTempl.append(-99.)
		ellg_perTempl.append(-99.)

		# Calculate the reduced model flux and calculate the final log-likleihood, knowing the optimised parameters.
		fluxSpecModelBG_red = []
		fluxSpecModelSG_red = []
		for i in range(0, Npoints_red-1):
			o = np.where((wavSpec >= wavSpecAnchor[i]) & (wavSpec < wavSpecAnchor[i+1]))[0]
			fluxSpecModelBG_red.append(np.mean(modelBG[o]))
			fluxSpecModelSG_red.append(np.mean(modelSG[o]))

		modelBG_red = np.concatenate([fluxSpecModelBG_red, modelBG_phot])
		modelSG_red = np.concatenate([fluxSpecModelSG_red, modelSG_phot])

		logl = lnpostfn_spec_noAGN(np.array([lnDust_perTempl[-1], lnPAH_perTempl[-1]]), \
								   np.array([Pdust, PPAH]), \
								   modelBG_red, modelSG_red, \
								   fluxFit_red, efluxFit_red, \
								   UL_red, wei_red, z, wavFit_red)
		logl_red = logl/float(len(modelBG_red))
		logl_perTempl.append(logl)

		# Flag for the used of an AGN. Here 0., since no AGN is accounted for.
		AGNon.append(0.)
		# save the number of parameters
		nParms.append(2)
		# save the name of the template
		tplName_perTempl.append(name_i)
		# Save teh total absorption at 9.7 microns.
		tau9p7_save.append(round(_tau9p7,3))

		# Fit including the full AGN model.
		ndim = 8 #NUmber of parms
		nwalkers = int(2. * ndim) # Number of walkers

		# Define the starting parms. Each of them has been normalised to a mean of zero to ease convergence.
		parms = np.zeros(shape=(nwalkers, ndim))
		parms[:,0] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm Dust
		parms[:,1] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm PAH
		parms[:,2] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm PL AGN
		parms[:,3] = np.random.uniform(low = -1., high = 0., size=nwalkers) # alpha1
		parms[:,4] = np.random.uniform(low = 0., high = 1., size=nwalkers) # alpha2
		parms[:,5] = np.random.uniform(low = -1., high = 1., size=nwalkers) # 10micron
		parms[:,6] = np.random.uniform(low = -1., high = 1., size=nwalkers) # 18micron
		parms[:,7] = np.random.uniform(low = -10., high = 10., size=nwalkers) # Break

		# Set the ensemble sample of Goodman&Weare and run the MCMC for Nmc steps
		sampler = EnsembleSampler(nwalkers, ndim, lnpostfn_spec_wAGN, args = (np.array([Pdust, PPAH, Ppl, Pslope1, Pslope2, Plsg, Pllg, Pbreak]), \
								  modelBG, modelSG, fluxFit, efluxFit, UL, wei, z, wavFit, np.zeros(len(wavFit))))
		sampler.run_mcmc(parms, Nmc, progress=bool(pgrbar))

		# Build the flat chain after burning 20% of the chain and thinning to every 10 values.
		chain = sampler.get_chain(discard=NburnIn, thin=10, flat=True)

		# Save the results
		#Dust continuum
		lnDust_perTempl.append(round(np.median(chain[:,0]),3))
		elnDust_perTempl.append(round(np.std(chain[:,0]),3))

		# PAH emission (set to -20. for empirircal templates)
		if np.sum(modelSG) > 0.:
			lnPAH_perTempl.append(round(np.median(chain[:,1]),3))
			elnPAH_perTempl.append(round(np.std(chain[:,1]),3))
		else:
			lnPAH_perTempl.append(-20.)
			elnPAH_perTempl.append(0.0)

		# Norm of the AGN continuum
		lnAGN_perTempl.append(round(np.median(chain[:,2]),3))
		elnAGN_perTempl.append(round(np.std(chain[:,2]),3))

		# slope of the first power law
		alpha1_perTempl.append(round(np.median(chain[:,3]),3))
		ealpha1_perTempl.append(round(np.std(chain[:,3]),3))

		# slope of the second power law
		alpha2_perTempl.append(round(np.median(chain[:,4]),3))
		ealpha2_perTempl.append(round(np.std(chain[:,4]),3))

		# silicate emission at 11 micron
		lsg_perTempl.append(round(np.median(chain[:,5]),3))
		elsg_perTempl.append(round(np.std(chain[:,5]),3))

		# silicate emisison at 18 micron
		llg_perTempl.append(round(np.median(chain[:,6]),3))
		ellg_perTempl.append(round(np.std(chain[:,6]),3))

		# cut off or position of the break
		coff_perTempl.append(round(np.median(chain[:,7]),3))
		ecoff_perTempl.append(round(np.std(chain[:,7]),3))

		# Define the reduced model (AGN + galaxy) for the final comparison
		modelPL = 10**(lnAGN_perTempl[-1] + Ppl[0]) * AGNmodel_jit(wavFit, 15.*(1.+z), (coff_perTempl[-1] + Pbreak[0])*(1.+z), \
					   alpha1_perTempl[-1] + Pslope1[0], alpha2_perTempl[-1]+Pslope2[0], -3.5)
		modelSi11 = 10**(lsg_perTempl[-1] + Plsg[0]) * Gauss_jit(wavFit, np.log10(11.*(1.+z)), 0.05)
		modelSi18 = 10**(llg_perTempl[-1] + Pllg[0]) * Gauss_jit(wavFit, np.log10(18.*(1.+z)), 0.07)
		modelAGN = modelPL + modelSi11 + modelSi18

		modelAGNSpec_red = []
		for i in range(0, Npoints_red-1):
			o = np.where((wavSpec >= wavSpecAnchor[i]) & (wavSpec < wavSpecAnchor[i+1]))[0]
			modelAGNSpec_red.append(np.mean(modelAGN[o]))
		modelAGN_red = np.concatenate([modelAGNSpec_red, np.interp(wavPhot, wavFit, modelAGN)])

		# save the loglikelihood of the model knowing the optimised parameters
		theta = np.array([lnDust_perTempl[-1], lnPAH_perTempl[-1], lnAGN_perTempl[-1], alpha1_perTempl[-1], \
				 		  alpha2_perTempl[-1], lsg_perTempl[-1], llg_perTempl[-1], coff_perTempl[-1]])
		Pr = np.array([Pdust, PPAH, Ppl, Pslope1, Pslope2, Plsg, Pllg, Pbreak])
		logl = lnpostfn_spec_wAGN(theta, \
								  Pr, \
								  modelBG_red, modelSG_red, \
								  fluxFit_red, efluxFit_red, \
								  UL_red, wei_red, z, wavFit_red, modelAGN_red)

		logl_red = logl/float(len(modelBG_red))
		logl_perTempl.append(logl)

		# AGNon = 1 since AGN is accounted for
		AGNon.append(1.)
		# Save the number of parameters
		nParms.append(8)
		# Save the name of the galaxy template
		tplName_perTempl.append(name_i)
		# Save the total obscuration at 9.7 microns
		tau9p7_save.append(round(_tau9p7,3))

	# Find the best model and the Akaike weight
	bestModelInd, Awi = exctractBestModel(logl_perTempl, nParms, len(wavFit), corrected = False)
	bestModelFlag = np.zeros(len(AGNon))
	bestModelFlag[bestModelInd] = 1

	# Save the results in a table
	resDict = {'logNormGal_dust': np.array(lnDust_perTempl) + Pdust[0], 'elogNormGal_dust': np.array(elnDust_perTempl), \
			   'logNormGal_PAH': np.array(lnPAH_perTempl) + PPAH[0], 'elogNormGal_PAH': np.array(elnPAH_perTempl), \
			   'logNormAGN_PL': np.array(lnAGN_perTempl) + Ppl[0], 'elogNormAGN_PL': np.array(elnAGN_perTempl), \
			   'lBreak_PL': np.array(coff_perTempl) + Pbreak[0], 'elBreak_PL': np.array(ecoff_perTempl), \
			   'alpha1_PL': np.array(alpha1_perTempl) + Pslope1[0], 'ealpha1_PL': np.array(ealpha1_perTempl), \
			   'alpha2_PL': np.array(alpha2_perTempl) + Pslope2[0], 'ealpha2_PL': np.array(ealpha2_perTempl), \
			   'logNorm_Si11': np.array(lsg_perTempl) + Plsg[0], 'elogNorm_Si11': np.array(elsg_perTempl), \
			   'logNorm_Si18': np.array(llg_perTempl) + Pllg[0], 'elogNorm_Si18': np.array(ellg_perTempl), \
			   'logl': logl_perTempl, 'AGNon': AGNon, 'tplName': tplName_perTempl,\
			   'bestModelFlag': bestModelFlag, 'Aw': Awi, 'tau9p7': tau9p7_save}

	dfRes = pd.DataFrame(resDict)

	return dfRes


#################################################
#												#
#		PHOTOMETRIC VERSION OF THE FITTING		#
#												#
#################################################
@njit
def lnpostfn_photo_noAGN(theta, P, modelBG, modelSG, UL, fluxFit, efluxFit, wei, final_pass):

	"""
    This function calculates the log-likelihood between photometric data and model without AGN contribution.
    ------------
    :param theta: vector containing the parameters.
    :param P: vector containing the priors for each of the parameters.
    :param modelBG: model dust continuum template.
    :param modelSG: model PAH template.
    :param UL: vector contaning the upper-limits, if any.
    :param fluxFit: observed fluxes.
    :param efluxFit: uncertainties on the observed fluxes.
    :param wei: weighting of the data points.
    ------------
    :return logl: log-likelihood of the model knowing the parameters theta.
    """
	
	# set the log-likelihood to zero
	logl = 0
	
	if final_pass == 0.:
		# prior constrain on the normalisation of the dust continuum for galaxy
		logl += -0.5*(theta[0])**2./P[0][1]**2.

		# prior constrain on the normalisation of the PAH
		logl += -0.5*(theta[1])**2./P[1][1]**2.
		# constrain on 0.3 dex between the normalisation of the dust to that of PAHs
		logl += - 0.5*(theta[1] + P[1][0] - (0.24 + 0.86 * (theta[0] + P[0][0])))**2./0.1**2.

	# define the full model as PAH + dust continuum
	ymodel = 10**(theta[0] + P[0][0]) * modelBG + 10**(theta[1] + P[1][0]) * modelSG

	# calculate the log-likelihood for upper-limits
	o = np.where(UL == 1)[0]
	if len(o)>0:
		logl += loglUL(ymodel[o], fluxFit[o])

	# calculate the log-likelihood for detected values
	o = np.where(UL == 0)[0]
	if len(o) > 0:
		logl += logldet(ymodel[o], fluxFit[o], efluxFit[o], wei[o])

	return logl

@njit
def lnpostfn_photo_wAGN(theta, P, modelBG, modelSG, modelAGN, modelSi, UL, fluxFit, efluxFit,wei, final_pass):
	
	"""
    This function calculates the log-likelihood between photometric data and model that includes AGN contribution.
    ------------
    :param theta: vector containing the parameters.
    :param P: vector containing the priors for each of the parameters.
    :param modelBG: model dust continuum template.
    :param modelSG: model PAH template.
    :param modelAGN: model AGN continuum template.
    :param modelSi: model silicate emission template.
    :param UL: vector contaning the upper-limits, if any.
    :param fluxFit: observed fluxes.
    :param efluxFit: uncertainties on the observed fluxes.
    :param wei: weighting of the data points.
    ------------
    :return logl: log-likelihood of the model knowing the parameters theta.
    """
	# set the log-likelihood to zero
	logl = 0

	if final_pass == 0.:
		# prior constrain on the normalisation of the dust continuum for galaxy
		logl += -0.5*(theta[0])**2./P[0][1]**2.
		
		# prior constrain on the normalisation of the PAH
		logl += -0.5*(theta[1])**2./P[1][1]**2.
		# constrain on 0.3 dex between the normalisation of the dust to that of PAHs
		logl += -0.5*(theta[1] + P[1][0] - (0.24 + 0.86 * (theta[0] + P[0][0])))**2./0.1**2.

		# prior constrain on the normalisation of the AGN continuum
		logl += -0.5*(theta[2])**2./P[2][1]**2.

		# prior constrain on the normalisation of the silicate emission
		logl += -0.5*(theta[3])**2./P[3][1]**2.

	# define the full model as PAH + dust continuum + AGN continuum + silicate emission
	ymodel = 10**(theta[0] + P[0][0]) * modelBG + 10**(theta[1] + P[1][0]) * modelSG + 10**(theta[2] + P[2][0]) * modelAGN + 10**(theta[3] + P[3][0]) * modelSi

	# calculate the log-likelihood for upper-limits
	o = np.where(UL == 1)[0]
	if len(o)>0:
		logl += loglUL(ymodel[o], fluxFit[o])
	
	# calculate the log-likelihood for detected values
	o = np.where(UL == 0)[0]
	logl += logldet(ymodel[o], fluxFit[o], efluxFit[o], wei[o])

	return logl

def runSEDphotFit(lambdaObs, fluxObs, efluxObs, \
				  filters, \
				  z = 0.01, \
				  UL = [], \
				  Nmc = 10000, pgrbar = 1, \
				  NoSiem = False, \
				  Pdust = [10., 13], PPAH = [9., 3.], PnormAGN = [10., 3.], PSiEm = [10., 3.], \
				  templ = '', \
				  NOAGN = False):

	"""
    This function fits the observed photometric SED.
    ------------
    :param wav: observed wavelengths (in microns).
    :param fluxSpec: observed fluxes (in Jansky).
    :param efluxSpec: observed uncertainties on the fluxes (in Jansky).
    :param filters: name of the photometric filters to include in the fit.
    ------------
    :keyword z: redshift of the source. Default = 0.01.
    :keyword UL: vector of length Nphot, where Nphot is the number of photometric data. If any of the value is set to 1, the corresponding flux is set has an upper limit in the fit. Default = [].
    :keyword Nmc: numer of MCMC run. Default = 10000.
    :keyword pgrbar: if set to 1, display a progress bar while fitting the SED. Default = 1.
	:keyword NoSiem: if set to True, no silicate emission template is included in the fit. Default = False.
    :keyword Pdust: normal prior on the log-normalisation of the galaxy dust continuum template. Default = [10., 3.] ([mean, std dev]).
    :keyword PPAH: normal prior on the log-normalisation of the PAH template. Default = [9., 3.] ([mean, std dev]).
    :keyword PnormAGN: normal prior on the log-normalisation of the AGN template. Default = [10., 3.] ([mean, std dev]).
    :keyword PSiem: normal prior on the log-normalisation of the silicate emission template. Default = [10., 3.] ([mean, std dev]).
    :keyword templ: normal prior on the log-normalisation of the silicate emission template. Default = [10., 3.] ([mean, std dev]).
    :keyword NOAGN: if set to True, fits are ran with SF templates only (i.e. no AGN emission is accounted for). Default = False.
	------------
    :return dfRes: dataframe containing the results of all the possible fits.
    """

	# If no templates are passed, open the Bernhard+20 templates.
	if len(templ) == 0:
		path = os.path.dirname(iragnsep.__file__)
		templ = pd.read_csv(path+'/iragnsep_templ.csv')

	# Extract the name of the templates
	keys = templ.keys().values
	nameTempl_gal = []
	nameTempl_PAH = []
	nameTempl_AGN = []
	nameTempl_Siem = []
	for key in keys:
		if str(key).startswith('gal'):
			if str(key).endswith('PAH') == False:
				nameTempl_gal.append(key)
			else:
				nameTempl_PAH.append(key)
		if str(key).startswith('AGN'):
			if str(key).endswith('Siem'):
				nameTempl_Siem.append(key)
			else:
				nameTempl_AGN.append(key)

	# Test that we have template for everything (if no galaxy then it crashes)
	if len(nameTempl_gal) == 0:
		raise ValueError('The galaxy template does not exist. The name of the column defining nuLnu for the galaxy template needs to start with "gal".')

	# define the wavelengths
	try:
		wavTempl = templ['lambda_mic'].values
	except:
		raise ValueError('Rename the wavelengths column of the template "lambda_mic".')

	# calculate the uncertainties on the fluxes
	unc = efluxObs/fluxObs

	# define the weighting accordingly
	if len(lambdaObs) > 1.:
		wei = np.ones(len(lambdaObs))
		wei_i = np.ones(len(lambdaObs))
		wei_final = np.ones(len(lambdaObs))
		o = np.where(unc > 0.)[0]
		wei_i[o] = 1./unc[o]**2./np.sum(1./unc[o]**2.)
		wei[o] = (1./wei_i[o])/np.sum(1./wei_i[o])
		wei_final = wei*np.gradient(lambdaObs)
	else:
		wei = np.ones(len(lambdaObs))
		wei_final = np.ones(len(lambdaObs))

	# Define a vectors of zeros if no upper limts are passed.
	if len(UL) != len(lambdaObs):
		UL = np.zeros(len(lambdaObs))

	o = np.where(UL == 0.)[0]
	if len(o) == 0.:
		UL = np.zeros(len(lambdaObs))
		efluxObs = fluxObs*0.1

	# Define the free parameters
	# Norm AGN continuum
	lnAGN_perTempl = []
	elnAGN_perTempl= []

	# Norm silicate emission
	lnSi_perTempl = []
	elnSi_perTempl = []

	# Norm dust continuum
	lnDust_perTempl = []
	elnDust_perTempl = []

	# Norm PAHs
	lnPAH_perTempl = []
	elnPAH_perTempl = []

	# final loglikelihood of the model
	logl_perTempl = []
	# name of the AGN and galaxy template
	tplNameGal_perTempl = []
	tplNameAGN_perTempl = []
	# if AGN is accounted for (1) or not (0)
	AGNon = []
	# Number of parameters in the model
	nParms = []

	# We loop over the 6 galaxy templates, first fitting only galaxy templates and then including the AGN templates.
	for name_i in nameTempl_gal:
		assert isinstance(name_i, str), "The list nameTempl requests strings as it corresponds to the names" + \
										" given to the various templates of galaxies to use for the fit."

		if pgrbar == 1:
			print("****************************************")
			print("  Fit of "+name_i+" as galaxy template  ")
			print("****************************************")

		# calculate the synthetic photometry of the Dust continuum template at wavelengths lambdaObs.
		nuLnuBGTempl = templ[name_i].values
		# Convert to photometric flux
		SEDgen = modelToSED(wavTempl, nuLnuBGTempl, z)
		modelBG = []
		for filt in filters:
			modelBG.append(getattr(SEDgen, filt)())

		modelBG = np.array(modelBG)

		# calculate the synthetic photometry of the PAH template at wavelengths lambdaObs.
		# When the empirical template is used, simply set a vector of zeros, so no separate contributions are accounted for as PAH emission.
		if name_i.endswith('dust') == False:
			modelSG = modelBG * 0.
			emodelSG = modelBG * 0.
		else:
			nuLnuSGTempl = templ['gal_PAH'].values
			SEDgen = modelToSED(wavTempl, nuLnuSGTempl, z)
			modelSG = []
			for filt in filters:
				modelSG.append(getattr(SEDgen, filt)())

			modelSG = np.array(modelSG)

		# Perform the fit without the AGN contribution
		ndim = 2 # Number of free params
		nwalkers = int(2. * ndim) # Number of walkers

		# Define the parameters as flat distributions between -1 and 1. (We normalised to zero each parameters to ease convergence).
		parms = np.zeros(shape=(nwalkers, ndim))
		parms[:,0] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm Dust
		parms[:,1] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm PAH

		# Set the ensemble sampler of Goodman&Weare and run the MCMC for Nmc steps
		sampler = EnsembleSampler(nwalkers, ndim, lnpostfn_photo_noAGN, args = (np.array([Pdust, PPAH]), modelBG, modelSG, UL, fluxObs, efluxObs,wei,0.))
		sampler.run_mcmc(parms, Nmc, progress=bool(pgrbar))

		# Build the flat chain, after burning 20% of the chain and thinning to every 10 values.
		NburnIn = int(0.2 * Nmc)
		chain = sampler.get_chain(discard=NburnIn, thin=10, flat=True)

		# Save the best fit parameters. Median of the posterior is taken as the best fit parameter and the standard deviation as 1sigma uncertainties.
		# Norm dust continuum
		lnDust_perTempl.append(round(np.median(chain[:,0]),3))
		elnDust_perTempl.append(round(np.std(chain[:,0]),3))

		# Norm PAH emission (or -20. when the empirical template is used).
		lnPAH_perTempl.append(round(np.median(chain[:,1]),3))
		elnPAH_perTempl.append(round(np.std(chain[:,1]),3))

		# Calulate the final loglikelihood of the model, using the best fit parameters.
		logl_perTempl.append(round(lnpostfn_photo_noAGN(np.array([lnDust_perTempl[-1], lnPAH_perTempl[-1]]), np.array([Pdust, PPAH]), \
								   modelBG, modelSG, UL, fluxObs, efluxObs,wei_final,1.),3))

		if np.sum(modelSG) == 0.:
			lnPAH_perTempl[-1] = -20.
			elnPAH_perTempl[-1] = 0.0
		
		# Norm on the silicate emission. -99. here since no AGN is accounted for.
		lnSi_perTempl.append(-99.)
		elnSi_perTempl.append(-99.)

		# Norm on the AGN template. -99. here since no AGN is accounted for.
		lnAGN_perTempl.append(-99.)
		elnAGN_perTempl.append(-99.)

		# No AGN in this fits, so AGNon set to zero
		AGNon.append(0.)
		# Save the numbers of parameters used in the fit, i.e. 2.
		nParms.append(2.)
		# Save the name of the template for the galaxy.
		tplNameGal_perTempl.append(name_i)
		# No AGN templates used in this fit, so name of the AGN template is set to 'N/A'.
		tplNameAGN_perTempl.append(str('N/A'))

		if NOAGN != True:

			# Fit including the AGN. Loop over the two templates AGN A and AGN B.
			for AGN_i in nameTempl_AGN:

				# calculate the synthetic photometry of the AGN template at wavelengths lambdaObs.
				nuLnu_AGN = templ[AGN_i].values
				# generate the photometry
				SEDgen = modelToSED(wavTempl, nuLnu_AGN, z)
				modelAGN = []
				for filt in filters:
					modelAGN.append(getattr(SEDgen, filt)())

				modelAGN = np.array(modelAGN)

				# calculate the synthetic photometry of the silicate template at wavelengths lambdaObs.
				# If no silicate emission is considered, set a vector of zeros so that no silicate emisison is account for.
				modelSiem = []
				if NoSiem == False:
					nuLnu_Siem = templ[nameTempl_Siem].values.flatten()
					SEDgen = modelToSED(wavTempl, nuLnu_Siem, z)
					for filt in filters:
						modelSiem.append(getattr(SEDgen, filt)())

					modelSiem = np.array(modelSiem)
				else:
					modelSiem = modelAGN * 0.

				# Perform the fit with the AGN contribution
				ndim = 4 # Number of parmameters
				nwalkers = int(2. * ndim) # Number of walkers

				# Define the starting parameters as flat distributions between -1 and 1. We normalised each parameter to zero to each convergence.
				parms = np.zeros(shape=(nwalkers, ndim))
				parms[:,0] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm Dust
				parms[:,1] = np.random.uniform(low = -1., high = 1., size=nwalkers) # norm PAH
				parms[:,2] = np.random.uniform(low = -1., high = 1., size=nwalkers) # normAGN
				parms[:,3] = np.random.uniform(low = -1., high = 1., size=nwalkers) # normSi

				# Set the ensemble sampler of Goodman&Weare and run the MCMC for Nmc steps.
				sampler = EnsembleSampler(nwalkers, ndim, lnpostfn_photo_wAGN, args = (np.array([Pdust, PPAH, PnormAGN, PSiEm]), \
										  modelBG, modelSG, modelAGN, modelSiem, UL, fluxObs, efluxObs,wei,0.))
				sampler.run_mcmc(parms, Nmc, progress=bool(pgrbar))

				# Build the flat chain, after burning 20% of the chain and thinning to every 10 values in the chain.
				chain = sampler.get_chain(discard=NburnIn, thin=10, flat=True)

				# Save the best fit parameters. Median of the posterior is taken as the best fit parameter and the standard deviation as 1sigma uncertainties.
				# Dust Normalisation
				lnDust_perTempl.append(round(np.median(chain[:,0]),3))
				elnDust_perTempl.append(round(np.std(chain[:,0]),3))

				# PAH normalisation
				lnPAH_perTempl.append(round(np.median(chain[:,1]),3))
				elnPAH_perTempl.append(round(np.std(chain[:,1]),3))
				
				# AGN continuum Norm
				lnAGN_perTempl.append(round(np.median(chain[:,2]),3))
				elnAGN_perTempl.append(round(np.std(chain[:,2]),3))

				# Si emission Norm
				lnSi_perTempl.append(round(np.median(chain[:,3]),3))
				elnSi_perTempl.append(round(np.std(chain[:,3]),3))
				
				# Numbers of params in the model
				nParms.append(4.)
	 
				# AGN accounted for in this case, so AGNon = 1
				AGNon.append(1.)

				# Name of the galaxy template
				tplNameGal_perTempl.append(name_i)

				# Name of the AGN template
				tplNameAGN_perTempl.append(AGN_i)

				# loglikelihood of the model
				logl_perTempl.append(round(lnpostfn_photo_wAGN(np.array([lnDust_perTempl[-1], lnPAH_perTempl[-1], lnAGN_perTempl[-1], \
										   lnSi_perTempl[-1]]), np.array([Pdust, PPAH, PnormAGN, PSiEm]), modelBG, modelSG, modelAGN, \
										   modelSiem, UL, fluxObs, efluxObs,wei_final,1.),3))

				if np.sum(modelSG) == 0.:
					lnPAH_perTempl[-1] = -20.
					elnPAH_perTempl[-1] = 0.0

				if NoSiem == True:
					lnSi_perTempl[-1] = -20.
					elnSi_perTempl[-1] = 0.0

	# Find the best model and the Akaike weight amongst all the 18 possible fits by comparing their final loglikelihood
	bestModelInd, Awi = exctractBestModel(logl_perTempl, nParms, len(lambdaObs), corrected = True)
	bestModelFlag = np.zeros(len(AGNon))
	bestModelFlag[bestModelInd] = 1

	# Save the results in a table
	resDict = {'logNormGal_dust': np.array(lnDust_perTempl) + Pdust[0], 'elogNormGal_dust': np.array(elnDust_perTempl), \
			   'logNormGal_PAH': np.array(lnPAH_perTempl) + PPAH[0], 'elogNormGal_PAH': np.array(elnPAH_perTempl), \
			   'logNormAGN': np.array(lnAGN_perTempl) + PnormAGN[0], 'elogNormAGN': np.array(elnAGN_perTempl), \
			   'logNormSiem': np.array(lnSi_perTempl) + PSiEm[0], 'elogNormSiem': np.array(elnSi_perTempl), \
			   'logl': logl_perTempl, 'AGNon': AGNon, 'tplName_gal': tplNameGal_perTempl, 'tplName_AGN': tplNameAGN_perTempl,\
			   'bestModelFlag': bestModelFlag, 'Aw': Awi}

	dfRes = pd.DataFrame(resDict)

	return dfRes


import numpy as np
import os
import iragnsep
import glob

from astropy.cosmology import WMAP9 as cosmo
from astropy import units as u
from astropy import constants as const
from scipy import integrate
from scipy.interpolate import UnivariateSpline, interp1d
from scipy.constants import h,k
from numba import njit

c = const.c.value
Lsun = const.L_sun.value

def get_prop(df, echar, z = 0.01, specOn = True, templ = ''):

	"""
    This function calculates the IR properties of the AGN and their hosts.
    ------------
    :param df: data-frame containing the results from the fits (i.e. optimised parameters) as returned by SEDanalysis.
    :param echar: characteristic uncertainties to calculate the uncertainties on the parameters.
    ------------
    :keyword z: redshift
	:keyword specOn: set to True if the data contain a spectrum in addition to the photometry.
    :keyword templ: set the templates that have been used in the fits.
    ------------
    :return loglum_hostIR: the host IR (8--1000microns) log-luminosity free of AGN contamination (Lsun).
    :return eloglum_hostIR: uncertainties on loglum_hostIR.
	:return loglum_hostMIR: the host IR (5--35microns) log-luminosity free of AGN contamination (Lsun).
    :return eloglum_hostIR: uncertainties on loglum_hostMIR.
	:return loglum_hostFIR: the host IR (40--1000microns) log-luminosity free of AGN contamination (Lsun).
    :return eloglum_hosFIR: uncertainties on loglum_hostFIR.
	:return loglum_AGNIR: the AGN IR log-luminosity free of host contamination (Lsun).
	:return loglum_AGNMIR: the AGN MIR log-luminosity free of host contamination (Lsun).
	:return loglum_AGNFIR: the AGN FIR log-luminosity free of host contamination (Lsun).
	:return AGNfrac_IR: the AGN fraction in the IR.
	:return AGNfrac_MIR: the AGN fraction in the MIR.
	:return AGNfrac_FIR: the AGN fraction in the FIR.
	:return SFR: the SFR of the galaxy free of AGN contamination.
	:return eSFR: the uncertainties on SFR.
	:return wSFR: the SFR of the galaxy free of AGN contamination weighted by its Akaike weight.
	:return ewSFR: the uncertainties on wSFR.
    """

	if len(templ) == 0:
		path = os.path.dirname(iragnsep.__file__)
		templ = pd.read_csv(path+'/iragnsep_templ.csv')

	# Extract the name of the templates
	keys = templ.keys().values
	nameTempl_gal = []
	nameTempl_AGN = []
	nameTempl_PAH = []
	nameTempl_Siem = []
	for key in keys:
		if str(key).startswith('gal'):
			if str(key).endswith('PAH') == False:
				nameTempl_gal.append(key)
			else:
				nameTempl_PAH.append(key)
		if str(key).startswith('AGN'):
			if str(key).endswith('Siem'):
				nameTempl_Siem.append(key)
			else:
				nameTempl_AGN.append(key)

	# Test that we have template for everything (if no galaxy then it crashes)
	if len(nameTempl_gal) == 0:
		raise ValueError('The galaxy template does not exist. The name of the column defining nuLnu for the galaxy template needs to start with "gal".')
	if len(nameTempl_AGN) == 0:
		print('Warning: The template for AGN is empty. The name of the column defining nuLnu for the AGN templates needs to start with "AGN".')

	# define the wavelengths
	try:
		wavTempl = templ['lambda_mic'].values
	except:
		raise ValueError('Rename the wavelengths column of the template "lambda_mic".')

	nu = c/wavTempl/1e-6 #Hz
	o_IR = np.where((nu > c/1000./1e-6) & (nu < c/8./1e-6))[0][::-1]
	o_MIR = np.where((nu > c/35./1e-6) & (nu < c/5./1e-6))[0][::-1]
	o_FIR = np.where((nu > c/1000./1e-6) & (nu < c/40./1e-6))[0][::-1]
	dMpc = cosmo.luminosity_distance(z).value
	dmeter = dMpc*u.Mpc.to(u.m)
	d2z = dmeter**2./(1.+z) # K correction=>https://ned.ipac.caltech.edu/level5/Sept02/Hogg/Hogg2.html
	JyToLsun = 1e-26 * 4. * np.pi * d2z/Lsun

	loglum_hostIR = []
	eloglum_hostIR = []
	loglum_hostMIR = []
	eloglum_hostMIR = []
	loglum_hostFIR = []
	eloglum_hostFIR = []
	loglum_AGNIR = []
	loglum_AGNMIR = []
	loglum_AGNFIR = []

	if specOn == True:

		for i in range(0, len(df)):

			obj = df.iloc[i]

			#host IR luminosity
			normDust = 10**obj['logNormGal_dust']
			enormDust = echar * np.log(10)
			
			# enormDust = obj['elogNormGal_dust'] * np.log(10)
			nuLnuDust = normDust * templ[obj['tplName']].values
			enuLnuDust = enormDust * nuLnuDust

			normPAH = 10**obj['logNormGal_PAH']
			# enormPAH = obj['elogNormGal_PAH'] * np.log(10)
			enormPAH = echar * np.log(10)
			nuLnuPAH = normPAH * templ['gal_PAH'].values
			enuLnuPAH = enormPAH * nuLnuPAH

			nuLnuGal = nuLnuDust + nuLnuPAH
			enuLnuGal = np.sqrt(enuLnuDust ** 2. + enuLnuPAH ** 2.)
			LnuGal = nuLnuGal/nu
			eLnuGal = enuLnuGal/nu

			loglum_hostIR.append(round(np.log10(np.trapz(LnuGal[o_IR], nu[o_IR], dx = np.gradient(nu[o_IR]))),3)) #Lsun
			elum_hostIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_IR[1:-2]], nu[o_IR[2:-1]], nu[o_IR[3:]], eLnuGal[o_IR[2:-1]], eLnuGal[o_IR[3:]]):
				elum_hostIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostIR = np.sqrt(elum_hostIR_i)
			eloglum_hostIR.append(round(0.434 * elum_hostIR/10**np.array(loglum_hostIR[-1]),3))
			
			loglum_hostMIR.append(round(np.log10(np.trapz(LnuGal[o_MIR], nu[o_MIR], dx = np.gradient(nu[o_MIR]))),3)) #Lsun
			elum_hostMIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_MIR[1:-2]], nu[o_MIR[2:-1]], nu[o_MIR[3:]], eLnuGal[o_MIR[2:-1]], eLnuGal[o_MIR[3:]]):
				elum_hostMIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostMIR = np.sqrt(elum_hostMIR_i)
			eloglum_hostMIR.append(round(0.434 * elum_hostMIR/10**np.array(loglum_hostMIR[-1]),3))
			
			loglum_hostFIR.append(round(np.log10(np.trapz(LnuGal[o_FIR], nu[o_FIR], dx = np.gradient(nu[o_FIR]))),3)) #Lsun
			elum_hostFIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_FIR[1:-2]], nu[o_FIR[2:-1]], nu[o_FIR[3:]], eLnuGal[o_FIR[2:-1]], eLnuGal[o_FIR[3:]]):
				elum_hostFIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostFIR = np.sqrt(elum_hostFIR_i)
			eloglum_hostFIR.append(round(0.434 * elum_hostFIR/10**np.array(loglum_hostFIR[-1]),3))
			
			if obj['AGNon'] == 1:
				modelAGN1 = 10**obj['logNorm_Si11'] * Gauss_jit(wavTempl*(1.+z), np.log10(11.*(1.+z)), 0.05)
				modelAGN2 = 10**obj['logNorm_Si18'] * Gauss_jit(wavTempl*(1.+z), np.log10(18.*(1.+z)), 0.07)
				modelAGN3 = 10**obj['logNormAGN_PL'] * AGNmodel_jit(wavTempl*(1.+z), 15.*(1.+z), obj['lBreak_PL']*(1.+z), obj['alpha1_PL'], obj['alpha2_PL'], -3.5)

				LnuAGN = (modelAGN1 + modelAGN2 + modelAGN3) * JyToLsun
				
				loglum_AGNIR.append(round(np.log10(np.trapz(LnuAGN[o_IR], nu[o_IR], dx = np.gradient(nu[o_IR]))),3)) #Lsun
				loglum_AGNMIR.append(round(np.log10(np.trapz(LnuAGN[o_MIR], nu[o_MIR], dx = np.gradient(nu[o_MIR]))),3)) #Lsun
				loglum_AGNFIR.append(round(np.log10(np.trapz(LnuAGN[o_FIR], nu[o_FIR], dx = np.gradient(nu[o_FIR]))),3)) #Lsun
			else:
				loglum_AGNIR.append(0.0)
				loglum_AGNMIR.append(0.0)
				loglum_AGNFIR.append(0.0)
	else:
		for i in range(0, len(df)):

			obj = df.iloc[i]

			#host IR luminosity
			normDust = 10**obj['logNormGal_dust']
			enormDust = obj['elogNormGal_dust'] * np.log(10)
			nuLnuDust = normDust * templ[obj['tplName_gal']].values
			enuLnuDust = enormDust * nuLnuDust

			normPAH = 10**obj['logNormGal_PAH']
			enormPAH = obj['elogNormGal_PAH'] * np.log(10)
			nuLnuPAH = normPAH * templ['gal_PAH'].values
			enuLnuPAH = enormPAH * nuLnuPAH

			nuLnuGal = nuLnuDust + nuLnuPAH
			enuLnuGal = np.sqrt(enuLnuDust ** 2. + enuLnuPAH ** 2.)
			LnuGal = nuLnuGal/nu
			eLnuGal = enuLnuGal/nu

			loglum_hostIR.append(round(np.log10(np.trapz(LnuGal[o_IR], nu[o_IR], dx = np.gradient(nu[o_IR]))),3)) #Lsun
			elum_hostIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_IR[1:-2]], nu[o_IR[2:-1]], nu[o_IR[3:]], eLnuGal[o_IR[2:-1]], eLnuGal[o_IR[3:]]):
				elum_hostIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostIR = np.sqrt(elum_hostIR_i)
			eloglum_hostIR.append(round(0.434 * elum_hostIR/10**np.array(loglum_hostIR[-1]),3))
			
			loglum_hostMIR.append(round(np.log10(np.trapz(LnuGal[o_MIR], nu[o_MIR], dx = np.gradient(nu[o_MIR]))),3)) #Lsun
			elum_hostMIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_MIR[1:-2]], nu[o_MIR[2:-1]], nu[o_MIR[3:]], eLnuGal[o_MIR[2:-1]], eLnuGal[o_MIR[3:]]):
				elum_hostMIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostMIR = np.sqrt(elum_hostMIR_i)
			eloglum_hostMIR.append(round(0.434 * elum_hostMIR/10**np.array(loglum_hostMIR[-1]),3))
			
			loglum_hostFIR.append(round(np.log10(np.trapz(LnuGal[o_FIR], nu[o_FIR], dx = np.gradient(nu[o_FIR]))),3)) #Lsun
			elum_hostFIR_i = 0
			for nu_i, nu_j, nu_k, eLnuGal_j, eLnuGal_k in zip(nu[o_FIR[1:-2]], nu[o_FIR[2:-1]], nu[o_FIR[3:]], eLnuGal[o_FIR[2:-1]], eLnuGal[o_FIR[3:]]):
				elum_hostFIR_i += 1./4. * (nu_j - nu_i)**2. * eLnuGal_j ** 2. + 1./4. * (nu_k - nu_j)**2. * eLnuGal_k ** 2.
			elum_hostFIR = np.sqrt(elum_hostFIR_i)
			eloglum_hostFIR.append(round(0.434 * elum_hostFIR/10**np.array(loglum_hostFIR[-1]),3))

			if obj['AGNon'] == 1:
				#AGN IR luminosity
				normAGN = 10**obj['logNormAGN']
				nuLnuAGN = normAGN * templ[obj['tplName_AGN']].values

				normSi = 10**obj['logNormSiem']
				nuLnuSi = normSi * templ[nameTempl_Siem].values.flatten()

				LnuAGN = (nuLnuAGN + nuLnuSi)/nu
				
				loglum_AGNIR.append(round(np.log10(np.trapz(LnuAGN[o_IR], nu[o_IR], dx = np.gradient(nu[o_IR]))),3)) #Lsun
				loglum_AGNMIR.append(round(np.log10(np.trapz(LnuAGN[o_MIR], nu[o_MIR], dx = np.gradient(nu[o_MIR]))),3)) #Lsun
				loglum_AGNFIR.append(round(np.log10(np.trapz(LnuAGN[o_FIR], nu[o_FIR], dx = np.gradient(nu[o_FIR]))),3)) #Lsun
			else:
				loglum_AGNIR.append(0.0)
				loglum_AGNMIR.append(0.0)
				loglum_AGNFIR.append(0.0)

	#Ratio of luminosities
	loglum_hostIR = np.array(loglum_hostIR)
	eloglum_hostIR = np.array(eloglum_hostIR)
	loglum_hostMIR = np.array(loglum_hostMIR)
	eloglum_hostMIR = np.array(eloglum_hostMIR)
	loglum_hostFIR = np.array(loglum_hostFIR)
	eloglum_hostFIR = np.array(eloglum_hostFIR)
	
	loglum_AGNIR = np.array(loglum_AGNIR)
	loglum_AGNMIR = np.array(loglum_AGNMIR)
	loglum_AGNFIR = np.array(loglum_AGNFIR)

	AGNfrac_IR = np.round(10**loglum_AGNIR/(10**loglum_hostIR + 10**loglum_AGNIR),2)
	o = np.where(loglum_AGNIR == 0.)[0]
	AGNfrac_IR[o] = 0.

	AGNfrac_MIR = np.round(10**loglum_AGNMIR/(10**loglum_hostMIR + 10**loglum_AGNMIR),2)
	o = np.where(loglum_AGNMIR == 0.)[0]
	AGNfrac_MIR[o] = 0.

	AGNfrac_FIR = np.round(10**loglum_AGNFIR/(10**loglum_hostFIR + 10**loglum_AGNFIR),2)
	o = np.where(loglum_AGNFIR == 0.)[0]
	AGNfrac_FIR[o] = 0.

	SFR = np.round(1.09e-10 * 10**loglum_hostIR,3) #Chabrier
	eSFR = np.round(1.09e-10 * 10**loglum_hostIR * np.log(10) * eloglum_hostIR,3)
	wSFR = np.round(1.09e-10 * 10**loglum_hostIR * df['Aw'].values,3)
	ewSFR = np.round(1.09e-10 * 10**loglum_hostIR * df['Aw'].values * np.log(10) * eloglum_hostIR,3)

	return loglum_hostIR, eloglum_hostIR, \
		   loglum_hostMIR, eloglum_hostMIR, \
		   loglum_hostFIR, eloglum_hostFIR, \
		   loglum_AGNIR, loglum_AGNMIR, loglum_AGNFIR, \
		   AGNfrac_IR, AGNfrac_MIR, AGNfrac_FIR, SFR, eSFR, wSFR, ewSFR

def basictests(wavSpec, fluxSpec, efluxSpec, wavPhot, fluxPhot, efluxPhot, filters, z, specOn = True):

	"""
    This function runs some basic tests prior to run the main fitting code.
    ------------
    :param wavSpec: observed wavelengths for the spectrum (in microns).
    :param fluxSpec: observed fluxes for the spectrum (in Jansky).
    :param efluxSpec: observed uncertainties on the fluxes for the spectrum (in Jansky).
    :param wavPhot: observed wavelengths for the photometry (in microns).
    :param fluxPhot: observed fluxes for the photometry (in Jansky).
    :param efluxPhot: observed uncertainties on the fluxes for the photometry (in Jansky).
    :param filters: name of the photometric filters to include in the fit.
    :param z: redshift.
    ------------
	:keyword specOn: set to True if the data contain a spectrum in addition to the photometry.
    ------------
    :return 0
    """

	if (len(wavPhot) != len(fluxPhot)) or (len(wavPhot) != len(efluxPhot)):
		raise ValueError("PHOTOMETRY ISSUE: Crashed because wavelengths, fluxes and uncertainties on the fluxes have different lengths.")
	if len(filters) != len(wavPhot):
		raise ValueError("FILTERS ISSUE: Crashed because the number of filters provided does not correspond to the number of photometry points.")
	if (any(fluxPhot<0) == True) or (any(wavPhot<0) == True):
		raise ValueError("PHOTOMETRY ISSUE: Crash caused by some negative values in the wavelengths, fluxes or uncertainties on the fluxes.")
	if (any(fluxPhot != fluxPhot) == True) or (any(efluxPhot != efluxPhot) == True) or (any(wavPhot != wavPhot) == True):
		raise ValueError("PHOTOMETRY ISSUE: Crash caused by some non-numerical values in the wavelengths, fluxes or uncertainties on the fluxes.")
	if specOn == True:
		#test that the length of wavelength is the same as the data
		if (len(wavSpec) != len(fluxSpec)) or (len(wavSpec) != len(efluxSpec)):
			raise ValueError("SPECTRUM ISSUE: Crashed because wavelengths, fluxes and uncertainties on the fluxes have different lengths.")
		#test that there are no negative values
		if (any(fluxSpec<0) == True) or (any(efluxSpec<0) == True) or (any(wavSpec<0) == True):
			raise ValueError("SPECTRUM ISSUE: Crash caused by some negative values in the wavelengths, fluxes or uncertainties on the fluxes.")
		#test that there are NAN
		if (any(fluxSpec != fluxSpec) == True) or (any(efluxSpec != efluxSpec) == True) or (any(wavSpec != wavSpec) == True):
			raise ValueError("SPECTRUM ISSUE: Crash caused by some non-numerical values in the wavelengths, fluxes or uncertainties on the fluxes.")

	#test if the filter exists
	path = os.path.dirname(iragnsep.__file__) + '/Filters/'
	files = [f for f in glob.glob(path + "*.csv")]
	count = -1
	for f in [path+f+"Filter.csv" for f in filters]:
		count += 1
		if (f in files) == False:
			raise ValueError(" \n The filter "+ str(filters[count]) + " does not exist. This version does not allow you to add some filters." + \
							 " Please get in touch with us to add the required filters (e.p.bernhard@sheffield.ac.uk). Available filters are:" +\
							 " IRAC1 , IRAC2 , IRAC3 , IRAC4 , WISE_W1 , WISE_W2 , WISE_W3 , WISE_W4 , IRAS12, IRAS60, IRAS100 , MIPS24, MIPS70," +\
							 " MIPS160 , PACS70, PACS100 , PACS160, SPIRE250ps , SPIRE350ps , SPIRE500ps")

	#Test if the redshift has been given by the user
	if z<0:
		zdefault = input('Warning: The redshift is set to the default value of 0.01. The keyword "z" allows you to indicate the redshift of the source.\n '+\
						 ' Press enter to continue, or type "exit" to abort.\n')
		if zdefault == "exit":
			exit()
		else:
			pass

	# Test that the wavelengths are in ascening order
	if len(wavPhot) > 1:
		dlambda = np.gradient(wavPhot)
		o = np.where(dlambda < 0.)[0]
		if len(o) > 0.:
			raise ValueError('PHOTOMETRY ISSUE: Wavelenghts need to be in ascending order.')

	if (specOn == True):
		dlambda = np.gradient(wavSpec)
		o = np.where(dlambda < 0.)[0]
		if len(o) > 0.:
			raise ValueError('SPECTRUM ISSUE: Wavelenghts need to be in ascending order.')

	if specOn == True:
		# Test if it can concatenate the Spectra aand the photometry
		try:
			wav = np.concatenate([wavSpec, wavPhot])
		except:
			raise ValueError("WAVELENGTHS: Spectral data cannot be concatenated to photometric data. Please check data.")
		try:
			flux = np.concatenate([fluxSpec, fluxPhot])
		except:
			raise ValueError("FLUXES: Spectral data cannot be concatenated to photometric data. Please check data.")
		try:
			eflux = np.concatenate([efluxSpec, efluxPhot])
		except:
			raise ValueError("UNCERTAINTIES ON THE FLUXES: Spectral data cannot be concatenated to photometric data. Please check data.")
	else:
		# Break as not enough data points anyway,
		if len(wavPhot) < 4:
			if len(wavPhot) == 3:
				restWav = wavPhot/(1.+z)
				NFIR = len(np.where(restWav>50.)[0])
				if (NFIR > 0.) & (NFIR != len(wavPhot)):
					pass
				else:
					raise ValueError('There is not enough data points to fit the model when compared to the number of degrees of freedom. It needs a minimum of' + \
								 ' 3 photometric points, with at least 1 FIR (i.e. rest-wavelength>60micron) flux, including upper-limits.')
	
			else:
				raise ValueError('There is not enough data points to fit the model when compared to the number of degrees of freedom. It needs a minimum of' + \
								 ' 3 photometric points, with at least 1 FIR (i.e. rest-wavelength>60micron) flux, including upper-limits.')


def exctractBestModel(logl, k, n, corrected = True):

	"""
    This function extracts the best model and calculates the Akaike weights based on the log-likelihood returned by the fits.
    ------------
    :param logl: log-likelihood returned by the fits.
    :param k: number of free parameters.
    :param n: number of data points.
    ------------
	:keyword corrected: if set to True, calculates the corrected AIC for small number of data points.
    ------------
    :return bestModelInd: the index of the best model fit.
    :return Awi: Akaike weights of each of the models, with respect to the best model.
    """

	nkdif = np.array(n)-np.array(k)
	o = np.where(nkdif == 1)[0]
	if len(o) > 0:
		corrected = False

	if corrected == True:
		AIC = 2*np.array(k) - 2.*np.array(logl) + (2.*np.array(k)**2. + 2.*np.array(k))/(np.array(n)-np.array(k)-1.)
	else:
		AIC = 2*np.array(k) - 2.*np.array(logl)

	bestModelInd = np.where(AIC == np.min(AIC))[0]
	AICmin = AIC[bestModelInd]
	AwiNorm = np.sum(np.exp(-0.5 * (AIC-AICmin)))

	Awi = np.exp(-0.5 * (AIC-AICmin))/AwiNorm

	return bestModelInd, Awi

def nuLnuToFnu(spec_wav, nuLnu, z):

	"""
    This function calculates the observed flux from nuLnu.
    ------------
    :param spec_wav: rest-wavelengths (in microns).
    :param nuLnu: nuLnu.
    :param z: redshift.
    ------------
    :return Fnu: observed flux on Earth of the source located at redshift z (in Jansky).
    """

	dMpc = cosmo.luminosity_distance(z).value #Mpc
	dmeter = dMpc*u.Mpc.to(u.m)
	d2z = dmeter**2./(1.+z) # K correction=>https://ned.ipac.caltech.edu/level5/Sept02/Hogg/Hogg2.html

	# Derive the observed flux
	nu = c/spec_wav/1e-6 #Hz
	Lnu = nuLnu/nu*Lsun #W/Hz
	Fnu = Lnu/4./np.pi/d2z #W/Hz/m2

	return Fnu * 1e26 # Jy

def getFluxInFilt(filt_wav, filt_QE, spec_wav, nuLnu, z):

	"""
    This function calculates the synthetic flux in a given filter and at a given redshift from a source with luminosity nuLnu.
    ------------
    :param filt_wav: passband of the filter.
	:param filt_QE: quantum efficient of the filter.
 	:param spec_wav: rest-wavelengths (in microns).
    :param nuLnu: nuLnu.
    :param z: redshift.
    ------------
    :return flux_Obs: observed flux on Earth of the source located at redshift z with luminosity nuLnu (in Jansky).
    """

	norm = integrate.trapz(filt_QE, x=c/filt_wav/1e-6)

	Fnu_0 = nuLnuToFnu(spec_wav, nuLnu, z) #Flux received on Earth
	lambda_0 = spec_wav * (1. + z) # Wavelenght of emission
	Fnu_0filt = np.interp(filt_wav, lambda_0, Fnu_0) # Move the template to grab the redshifted flux
	flux_Obs = integrate.trapz(Fnu_0filt*filt_QE, x = c/filt_wav/1e-6)/norm # This is the flux received on Earth

	return flux_Obs

@njit
def logldet(ym, ydat, eydat, wei):

	"""
    This function calculates the log-likelihood of detected data.
    ------------
    :param ym: model values.
	:param ydat: observed values.
 	:param eydat: observed uncertaities.
    :param wei: weights of the data points.
    ------------
    :return logl: log-likelihood of the model.
    """

	sigma = 0.434 * eydat/ydat
	logl = np.sum((-0.5*(np.log10(ydat) - np.log10(ym))**2./sigma**2.)*wei)

	return logl

@njit
def erf_approx(x):
	"""
    This function calculates an approximation of the error function.
    ------------
    :param x: x values to which the error function is calculated at.
    ------------
    :return erf_approx_eval: approximate value of the error function.
    """

	t_erf = 1./(1. + 0.5*np.abs(x))
	tau_erf = t_erf*np.exp(-x**2. - 1.26551223 + 1.00002368*t_erf + 0.37409196*t_erf**2. + 0.09678418*t_erf**3.\
								  - 0.18628806*t_erf**4., + 0.27886807*t_erf**5. - 1.13520398*t_erf**6. + 1.48851587*t_erf**7. \
								  - 0.82215223*t_erf**8. + 0.17087277*t_erf**9.)
	for i, xnden in np.ndenumerate(x):
		if xnden >= 0:
			erf_approx_eval = 1. - tau_erf
		else:
			erf_approx_eval = tau_erf - 1.

	return erf_approx_eval[0]

@njit
def loglUL(ym, ydat):

	"""
    This function calculates the log-likelihood of undetected data.
    ------------
    :param ym: model values.
	:param ydat: observed values.
    ------------
    :return logl: log-likelihood of the model.
    """

	
	logl = 0
	x = (np.log10(ym) - np.log10(ydat))/np.sqrt(2.)/0.15
	t_erf = 1./(1. + 0.5*np.abs(x))
	tau_erf = t_erf*np.exp(-x**2. - 1.26551223 + 1.00002368*t_erf + 0.37409196*t_erf**2. + 0.09678418*t_erf**3.\
								  - 0.18628806*t_erf**4., + 0.27886807*t_erf**5. - 1.13520398*t_erf**6. + 1.48851587*t_erf**7. \
								  - 0.82215223*t_erf**8. + 0.17087277*t_erf**9.)
	for i, xnden in np.ndenumerate(x):
		if xnden >= 0:
			erf_approx = 1. - tau_erf
			logl += np.sum(np.log(1. - (0.5 * (1. + erf_approx))))
		else:
			erf_approx = tau_erf - 1.
			logl += np.sum(np.log(1. - (0.5 * (1. + erf_approx))))

	return logl


@njit
def Gauss_jit(x, mu, sigma):
	"""
    This function calculates a Gaussian normalised to its maximum.
    ------------
    :param x: x-values.
	:param mu: mean.
 	:param sigma: standard deviation.
    ------------
    :return Bnu: Gaussian evaluated at x, normalised to its maximum.
    """

	Bnu = np.exp(-(np.log10(x)-mu)**2./2./sigma/sigma)

	return Bnu/np.max(Bnu)

@njit
def AGNmodel_firstPL_jit(x, lambdab, alpha1, alpha2, s):
	
	"""
    This function calculates the first broken power-law of the AGN model.
    ------------
    :param x: x-values.
	:param lambdab: position of the break.
 	:param alpha1: slope of the first power-law.
 	:param alpha2: slope of the second power-law.
    :param s: sharpness parameter for the break between alpha1 and alpha2. 
    ------------
    :return Bnu: corresponding broken power law evaluated at x, and normalised at 10 microns.
    """
	Bnu = x**alpha1*(1. + (x/lambdab)**(abs(alpha2-alpha1)*s))**(np.sign(alpha2-alpha1)/s)
	BnuNorm = 10.**alpha1*(1. + (10./lambdab)**(abs(alpha2-alpha1)*s))**(np.sign(alpha2-alpha1)/s)

	return Bnu/BnuNorm

@njit
def AGNmodel_jit(x, lambdab1, lambdab2, alpha1, alpha2, alpha3):

	"""
    This function calculates the second broken power-law of the AGN model.
    ------------
    :param x: x-values.
	:param lambdab1: position of the break for the first broken power law.
	:param lambdab2: position of the break for the second broken power law.
 	:param alpha1: slope of the first power-law.
 	:param alpha2: slope of the second power-law.
 	:param alpha2: slope of the third power-law.
    ------------
    :return Bnu: corresponding double-broken power law evaluated at x, and normalised at 10 microns.
    """

	Bnu0 = AGNmodel_firstPL_jit(x, lambdab1, alpha1, alpha2,2.)
	Bnu1 = x**alpha2*(1. + (x/lambdab2)**(abs(alpha3-alpha2)*2.))**(np.sign(alpha3-alpha2)/2.)

	xJoint = (lambdab1 + lambdab2)/2.
	Bnu1Norm = Bnu1/np.interp(xJoint, x, Bnu1)*np.interp(xJoint, x, Bnu0)

	Bnu = Bnu0
	o = np.where(x >= xJoint)[0]
	Bnu[o] = Bnu1Norm[o]

	return Bnu

def KVTextinctCurve(lambda_obs):

	"""
    This function calculates the extinction curve from Kemper et al. (2004).
    ------------
    :param lambda_obs: observed wavelengths.
    ------------
    :return tau_return: absorption coefficient tau per wavelength.
    """

	lambda_mic = np.arange(1., 1000., 0.01)
	kvt_wav = 8.0, 8.2, 8.4,  8.6,  8.8,  9.0,  9.2,  9.4, 9.6, 9.7, 9.75, 9.8, 10.0, 10.2,\
			  10.4, 10.6, 10.8, 11.0, 11.2, 11.4,11.6, 11.8, 12.0, 12.2, 12.4, 12.6, 12.7
	kvt_prof= .06, .09, .16, .275, .415, .575, .755, .895, .98, .99, 1.00, .99, .94 ,  .83,\
			  .745, .655,  .58, .525,  .43,  .35, .27,  .20,  .13,  .09,  .06, .045, .04314 

	PSIlambda = np.zeros(len(lambda_mic))

	o = np.where((lambda_mic > 8) & (lambda_mic < 12.7))[0]
	if len(o) > 0:
		PSIlambda[o] = np.interp(lambda_mic[o], kvt_wav, kvt_prof)

	PSI8 = 0.06
	o = np.where(lambda_mic <= 8)[0]
	if len(o) > 0:
		PSIlambda[o] = PSI8*np.exp(2.03*(lambda_mic[o]-8.))

	o = np.where(lambda_mic >= 12.7)[0]
	if len(o) > 0:
		PSIlambda[o] = 0.247 * drude(lambda_mic[o], 0.4, 18.)

	PSI9p7 = np.interp(9.7, lambda_mic, PSIlambda)
	tau = (0.9*PSIlambda + 0.1*(9.7/lambda_mic)**1.7)/PSI9p7

	tau_return = np.zeros(len(lambda_obs))+1e-10
	o = np.where(lambda_obs < max(lambda_mic))[0]
	if len(o) > 0:
		tau_return[o] = np.interp(lambda_obs[o], lambda_mic, tau)
	return tau_return

# Drude profile
def drude(x, gamma_r, lambda_r, normed = True):
	"""
    This function calculates a Drude profile.
    ------------
    :param x: x-values.
    :param gamma_r: central wavelengths.
    :param lambda_r: fractional FWHM.
    ------------
    :keyword normed: if set to True, normalise to the maximum value.
    ------------
    :return drudeVal: the Drude profile evaluated at x.
    """
	numerateur = gamma_r**2.
	denominateur = (x/lambda_r - lambda_r/x)**2. + gamma_r**2.
	drudeVal = numerateur/denominateur

	if normed == True:
		return drudeVal/np.max(drudeVal)
	else:
		return np.max(drudeVal), drudeVal

def obsC(wavRest, flux, eflux):
	"""
    This function calculates the total obscuration at 9.7micron.
    ------------
    :param wavRest: rest-wavelength.
    :param flux: observed fluxes.
    :param eflux: uncertainties on the observed fluxes.
    ------------
    :return _tau9p7: the total obscuration at 9.7micron.
    """
	loc1 = np.where((wavRest >= 5.) & (wavRest<=5.5))[0]
	loc2 = np.where((wavRest >= 14.7) & (wavRest<=15.4))[0]

	if (len(loc1) <1) & (len(loc2) < 1):
	 	raise Exception()
	k = 1

	# Get the flux in the anchored wavelengths
	wavAbs = np.concatenate((wavRest[loc1], wavRest[loc2]))
	fluxAbs = np.concatenate((flux[loc1], flux[loc2]))

	Npoints = float(len(loc1) + len(loc2))
	wei = np.zeros(len(wavAbs))
	wei[0:len(loc1)] = 1./(len(loc1)/Npoints)
	wei[len(loc1):len(loc1)+len(loc2)] = 1./(len(loc2)/Npoints)

	# Fit a 1-order spline
	o = np.where(np.diff(wavAbs) > 0.)[0]
	spl = UnivariateSpline(np.log10(wavAbs[o]), np.log10(fluxAbs[o]), k=k, w = wei[o])
	contFlux = 10**spl(np.log10(wavRest))

	# Test of AGN continuum is not miss-representing obscuration
	o = np.where((wavRest > 9.) & (wavRest < 10.3))[0]
	spl = UnivariateSpline(np.log10(wavRest[o]), np.log10(flux[o]), k=2)
	deepFlux = 10**spl(np.log10(wavRest[o]))

	Norm10 = np.interp(10., wavRest[o], contFlux[o])
	contFluxNorm = contFlux[o]/Norm10
	Norm10 = np.interp(10., wavRest[o], deepFlux)
	deepFluxNorm = deepFlux/Norm10

	diff = np.std(contFluxNorm/deepFluxNorm)
	if diff < 0.01:
		return np.array([-99.])

	# Measure the ratio between the true and the observed flux
	fluxTrue9p7 = np.interp(9.7, wavRest, contFlux)
	locDeep = np.where((wavRest >= 9.4) & (wavRest <= 10.0))[0]
	fluxObs9p7 = np.mean(flux[locDeep])
	S9p7 = fluxObs9p7/fluxTrue9p7

	if S9p7 >= 1.:
		return np.array([0.])
	else:
		tauvec = 10**np.arange(-5., 5., 0.01)
		S9p7vec = (1. - np.exp(-tauvec))/tauvec

		f = interp1d(S9p7vec, tauvec)
		_tau9p7 = np.array([f(S9p7)])

	return _tau9p7

	


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os

from .func import *
from .classes import *

from matplotlib.font_manager import FontProperties
font0 = FontProperties(family = 'serif', variant = 'small-caps', size = 22)

############################
## Plot all the results for all the possible fit. Tag the best model in blue.
############################
def plotFitSpec(df, wavSpec, fluxSpec, efluxSpec,\
				wavPhot, fluxPhot, efluxPhot, UL = np.array([]), pathFig = './', sourceName = 'NoName', templ = '', z = 0.01, saveRes = True):
	

	"""
    This function plots the results of the fits for data which contain spectral + photometry fluxes.
    ------------
    :param df: data-frame contaning the results of the fits (i.e. optimised parameters) as returned by the function SEDanalysis.runSEDspecFit.
    :param wavSpec: observed wavelengthts for the spectrum (in microns).
    :param fluxSpec: observed fluxes for the spectrum (in Jansky).
    :param efluxSpec: observed uncertainties on the fluxes for the spectrum (in Jansky).
    :param wavPhot: observed wavelengths for the photometry (in microns).
    :param fluxPhot: observed fluxes for the photometry (in Jansky).
    :param efluxPhot: observed uncertainties on the fluxes for the photometry (in Jansky).
    ------------
    :keyword z: redshift of the source. Default = 0.01.
    :keyword UL: vector of length Nphot, where Nphot is the number of photometric data. If any of the values is set to 1, the corresponding flux is set has an upper limit in the fit. Default = [].
  	:keyword sourceName: name of the source. Default = 'NoName'.
	:keyword pathFig: if saveRes is set to True, the figues showing the results of the fits will be saved at the location pathFig. Default = './'.
	:keyword saveRes: if set to True, the tables containing the results of the fits as well as the figures are saved. Default = True.
	:keyword templ: set the templates that have been used in the fits.
	------------
    :return 0
    """

    # concatenate the spectral and photo fluxes
	flux = np.concatenate([fluxSpec, fluxPhot])

	# define the upper limits if not.
	if len(UL) == 0.:
		UL = np.zeros(len(wavPhot))

	# define the templates to Bernhard+20 if not.
	if len(templ) == 0:
		path = os.path.dirname(iragnsep.__file__)
		templ = pd.read_csv(path+'/iragnsep_templ.csv')

	# Extract the name of the templates
	keys = templ.keys().values
	nameTempl_gal = []
	nameTempl_PAH = []
	for key in keys:
		if str(key).startswith('gal'):
			if str(key).endswith('PAH') == False:
				nameTempl_gal.append(key)
			else:
				nameTempl_PAH.append(key)

	# Test that we have templates for everything (if no galaxy then it crashes)
	if len(nameTempl_gal) == 0:
		raise ValueError('The galaxy template does not exist. The name of the column defining nuLnu for the galaxy template needs to start with "gal".')
	
	# define the wavelengths
	try:
		wavTempl = templ['lambda_mic'].values
	except:
		raise ValueError('Rename the wavelengths column of the template "lambda_mic".')

	# Open 2 figures, 1 for all the possible fits and 1 for the best weighted average model.
	fig1, axs1 = plt.subplots(2,6, sharex = True, sharey = True, figsize = (27, 12))
	fig1.subplots_adjust(hspace = 0.0, wspace=0., left  = 0.06, right = 0.99, bottom = 0.10, top = 0.95)
	axs1 = axs1.ravel()

	fig2, axs2 = plt.subplots(figsize = (11, 8))
	fig2.subplots_adjust(hspace = 0.0, wspace=0., left  = 0.15, right = 0.95, bottom = 0.15, top = 0.95)
	axs2.tick_params(axis='both', labelcolor='k', labelsize = 25, width = 1, size = 15, which = 'major', direction = 'inout')
	axs2.tick_params(axis='both', width = 1, size = 10, which = 'minor', direction = 'inout')

	# Loop on all the possible combinations of models
	count = 0
	for i in range(0, len(df)):
		obj = df.iloc[i]

		# Extract the optimised best parameters and their uncertainties
		normDust = 10**obj['logNormGal_dust']
		nuLnuDust = normDust * templ[obj['tplName']].values
		enuLnuDust = normDust * templ['e'+obj['tplName']].values

		normPAH = 10**obj['logNormGal_PAH']
		nuLnuPAH = normPAH * templ['gal_PAH'].values
		enuLnuPAH = normPAH * templ['egal_PAH'].values

		nuLnuGal = nuLnuDust + nuLnuPAH
		enuLnuGal = np.sqrt(enuLnuDust**2. + enuLnuPAH**2.)

		# Add the exctinction if measured
		_tau9p7 = obj['tau9p7']
		if _tau9p7 != -99.:
			tau = KVTextinctCurve(wavTempl) * _tau9p7
			# exctCorr = (1. - np.exp(-tau))/tau
			exctCorr = (1 - np.exp(-tau))/tau
		else:
			exctCorr = 1.

		# Calculate the observed flux, affected by attenuation
		FnuGal = nuLnuToFnu(wavTempl, nuLnuGal, z)*exctCorr
		eFnuGal = nuLnuToFnu(wavTempl, enuLnuGal, z)*exctCorr

		# Define the AGN contribution
		if obj['AGNon'] == 1:
			modelAGN1 = 10**obj['logNorm_Si11'] * Gauss_jit(wavTempl*(1.+z), np.log10(11.*(1.+z)), 0.05)
			modelAGN2 = 10**obj['logNorm_Si18'] * Gauss_jit(wavTempl*(1.+z), np.log10(18.*(1.+z)), 0.07)
			modelAGN3 =  10**obj['logNormAGN_PL'] * AGNmodel_jit(wavTempl*(1.+z), 15.*(1.+z), obj['lBreak_PL']*(1.+z), obj['alpha1_PL'], obj['alpha2_PL'], -3.5)

			FnuAGN = (modelAGN1 + modelAGN2 + modelAGN3)*exctCorr

			FnuTot = FnuGal + FnuAGN
		else:
			FnuTot = FnuGal

		# Figure with all the fits.
		o = np.where(UL < 1.)[0]
		axs1[count].errorbar(wavPhot[o]/(1.+z), fluxPhot[o], yerr = efluxPhot[o], fmt = '.', color = '#292F33', alpha = 1.0, label = 'Observed SED')
		o = np.where(UL > 0.)[0]
		axs1[count].errorbar(wavPhot[o]/(1.+z), fluxPhot[o], yerr = fluxPhot[o]/3., fmt = '.', color = '#292F33', alpha = 1.0, label = '_no_label_', uplims = True)
		axs1[count].errorbar(wavSpec/(1.+z), fluxSpec, yerr = efluxSpec, fmt = '.', color = '#292F33', alpha = 1.0, label = '_no_label_')

		if obj['AGNon'] == 1:
			axs1[count].plot(wavTempl, FnuGal, '--', color = '#66757F', label = 'Galaxy comp. ('+obj['tplName']+')')
			axs1[count].plot(wavTempl, FnuAGN, '-.', color = '#66757F', label = 'AGN comp. (full model)')
			if obj['bestModelFlag'] == 1.:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = '#55ACEE', ecolor = 'k', \
								linewidth = 3, elinewidth=0.5, label = 'Total (AGN + Galaxy)', errorevery = 3)
			else:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = 'k', ecolor = 'k', \
									linewidth = 1, elinewidth=0.5, label = 'Total (AGN + Galaxy)', errorevery = 3)
		else:
			if obj['bestModelFlag'] == 1.:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = '#55ACEE', ecolor = 'k', \
								linewidth = 3, elinewidth=0.5, label = 'Total ('+obj['tplName']+')', errorevery = 3)
			else:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = 'k', ecolor = 'k', \
									linewidth = 1, elinewidth=0.5, label = 'Total ('+obj['tplName']+')', errorevery = 3)


		axs1[count].set_xscale('log')
		axs1[count].set_yscale('log')
		axs1[count].set_xlim([3./(1.+z), 800./(1.+z)])
		axs1[count].set_ylim([min(flux)/10., max(flux)*20.])
		axs1[count].set_xlabel(r'$\lambda_{\rm rest}\ (\mu {\rm m})$', fontsize = 28)
		if (count == 0) or (count == 6):
			axs1[count].set_ylabel(r'Flux (Jy)', fontsize = 28)
		axs1[count].legend(frameon = False, loc = 'upper left', fontsize = 15)
		if round(obj['Aw']*100.) > 50.:
			axs1[count].text(10., min(flux)/6., 'Aw = '+str(round(obj['Aw']*100.))+'%', fontsize = 25, c = '#ff4d4d')
		else:
			axs1[count].text(10., min(flux)/6., 'Aw = '+str(round(obj['Aw']*100.))+'%', fontsize = 25, c = 'k')
		axs1[count].tick_params(axis='both', labelcolor='k', labelsize = 25, width = 1, size = 15, which = 'major', direction = 'inout')
		axs1[count].tick_params(axis='both', width = 1, size = 10, which = 'minor', direction = 'inout')

		count += 1

		# Best model Figure
		if obj['Aw'] > 0.1:
			if obj['AGNon'] == 1:
				axs2.plot(wavTempl, FnuTot, '-', color = 'k', \
						 linewidth = 1, label = obj['tplName']+' + AGN ('+str(round(obj['Aw']*100.))+'%)', alpha = obj['Aw']/2.)				
			else:
				axs2.plot(wavTempl, FnuTot, '-', color = 'k', \
						 linewidth = 1, label = obj['tplName']+' ('+str(round(obj['Aw']*100.))+'%)', alpha = obj['Aw']/2.)

		try:
			FnuTot_Aw += FnuTot * obj['Aw']
			eFnuTot_Aw += (eFnuGal*obj['Aw'])**2.
			FnuGal_Aw += FnuGal * obj['Aw']
			if obj['AGNon'] == 1:
				FnuAGN_Aw += FnuAGN * obj['Aw']
		except:
			FnuTot_Aw = FnuTot * obj['Aw']
			eFnuTot_Aw = (eFnuGal * obj['Aw'])**2.
			FnuGal_Aw = FnuGal * obj['Aw']
			if obj['AGNon'] == 1:
				FnuAGN_Aw = FnuAGN * obj['Aw']
			else:
				FnuAGN_Aw = FnuGal_Aw * 0.

	axs2.errorbar(wavTempl, FnuTot_Aw, yerr = np.sqrt(eFnuTot_Aw), fmt = '-', color = 'k', elinewidth = 0.5, linewidth = 3, label = 'Best weighted fit [Total]',\
			 ecolor = 'k', alpha = 0.4, errorevery = 3)
	axs2.plot(wavTempl, FnuGal_Aw, '--', color = '#E94B3C', \
						 linewidth = 2, label = 'Best weighted fit [Galaxy]', alpha = 0.7)
	axs2.plot(wavTempl, FnuAGN_Aw, '-.', color = '#6395F2', \
						 linewidth = 2, label = 'Best weighted fit [AGN]', alpha = 0.7)

	o = np.where(UL < 1.)[0]
	axs2.errorbar(wavPhot[o]/(1.+z), fluxPhot[o], yerr = efluxPhot[o], fmt = 'o', color = '#292F33', alpha = 0.9, label = 'Observed Photometry', mfc = 'none', mew = 1, ms = 10)

	o = np.where(UL > 0.)[0]
	axs2.errorbar(wavPhot[o]/(1.+z), fluxPhot[o], yerr = fluxPhot[o]/5., fmt = 'o', color = '#292F33', alpha = 0.9, label = '_no_label_', uplims = True, mfc = 'none', mew = 1, ms = 10)

	axs2.errorbar(wavSpec/(1.+z), fluxSpec, yerr = efluxSpec, fmt = '.', color = '#292F33', alpha = 0.9, label = 'Observed Spectrum')

	axs2.set_xscale('log')
	axs2.set_yscale('log')
	axs2.set_xlim([3./(1.+z), 800./(1.+z)])
	axs2.set_ylim([min(flux)/5., max(flux)*10.])
	axs2.set_xlabel(r'$\lambda_{\rm rest}\ (\mu {\rm m})$', FontProperties = font0)
	axs2.set_ylabel(r'Flux (Jy)', FontProperties = font0)
	axs2.legend(frameon = False, fontsize = 16, ncol = 2)

	# If saveRes is True, then save the figures to the location specified in pathFig.
	if saveRes == True:
		fig1.savefig(pathFig+sourceName+'_fitResAll_spec.pdf')
		fig2.savefig(pathFig+sourceName+'_fitResBM_spec.pdf')
	else:
		plt.show()
	plt.close('all')


############################
## Plot all the results for all the possible fit. Tag the best model in blue.
############################
def plotFitPhoto(df, wav, flux, eflux, UL = np.array([]), pathFig = './', sourceName = 'NoName', templ = '', z = 0.01, saveRes = True, NOAGN = False):
	
	"""
    This function plots the results of the fits for photometric data.
    ------------
    :param df: data-frame contaning the results of the fits (i.e. optimised parameters) as returned by the function SEDanalysis.runSEDspecFit.
    :param wav: observed wavelengthts (in microns).
    :param flux: observed fluxes (in Jansky).
    :param eflux: observed uncertainties on the fluxes (in Jansky).
    ------------
    :keyword z: redshift of the source. Default = 0.01.
    :keyword UL: vector of length Nphot, where Nphot is the number of photometric data. If any of the values is set to 1, the corresponding flux is set has an upper limit in the fit. Default = [].
  	:keyword sourceName: name of the source. Default = 'NoName'.
	:keyword pathFig: if saveRes is set to True, the figues showing the results of the fits will be saved at the location pathFig. Default = './'.
	:keyword saveRes: if set to True, the tables containing the results of the fits as well as the figures are saved. Default = True.
	:keyword templ: set the templates that have been used in the fits.
	------------
    :return 0
    """

    # Open the Bernhard+20 templates if not defined
	if len(templ) == 0:
		path = os.path.dirname(iragnsep.__file__)
		templ = pd.read_csv(path+'/iragnsep_templ.csv')

	# Extract the name of the templates
	keys = templ.keys().values
	nameTempl_gal = []
	nameTempl_PAH = []
	nameTempl_AGN = []
	nameTempl_Siem = []
	for key in keys:
		if str(key).startswith('gal'):
			if str(key).endswith('PAH') == False:
				nameTempl_gal.append(key)
			else:
				nameTempl_PAH.append(key)
		if str(key).startswith('AGN') == True:
			if str(key).endswith('Siem') == True:
				nameTempl_Siem.append(key)
			else:
				nameTempl_AGN.append(key)

	# Test that we have template for everything (if no galaxy then it crashes)
	if len(nameTempl_gal) == 0:
		raise ValueError('The galaxy template does not exist. The name of the column defining nuLnu for the galaxy template needs to start with "gal".')
	if len(nameTempl_AGN) == 0:
		raise ValueError('The AGN template does not exist. The name of the column defining nuLnu for the AGN template needs to start with "AGN".')

	# define the wavelengths
	try:
		wavTempl = templ['lambda_mic']
	except:
		raise ValueError('Rename the wavelengths column of the template "lambda_mic".')

	if NOAGN != True:
		# Open 2 figures, 1 for all the possible fits and 1 for the best weighted average model.
		fig1, axs1 = plt.subplots(3,6, sharex = True, sharey = True, figsize = (27, 18))
		fig1.subplots_adjust(hspace = 0.0, wspace=0., left  = 0.06, right = 0.99, bottom = 0.10, top = 0.95)
		axs1 = axs1.ravel()
	else:
		fig1, axs1 = plt.subplots(1,6, sharex = True, sharey = True, figsize = (27, 10))
		fig1.subplots_adjust(hspace = 0.0, wspace=0., left  = 0.06, right = 0.99, bottom = 0.10, top = 0.95)
		axs1 = axs1.ravel()


	fig2, axs2 = plt.subplots(figsize = (11, 8))
	fig2.subplots_adjust(hspace = 0.0, wspace=0., left  = 0.15, right = 0.95, bottom = 0.15, top = 0.95)
	axs2.tick_params(axis='both', labelcolor='k', labelsize = 25, width = 1, size = 15, which = 'major', direction = 'inout')
	axs2.tick_params(axis='both', width = 1, size = 10, which = 'minor', direction = 'inout')

	# Loop over all the model combinations of templates.
	count = 0
	for i in range(0, len(df)):
		obj = df.iloc[i]

		# Exrract the optimised parameters and their uncertainties
		normDust = 10**obj['logNormGal_dust']
		nuLnuDust = normDust * templ[obj['tplName_gal']].values
		enuLnuDust = normDust * templ['e'+obj['tplName_gal']].values

		normPAH = 10**obj['logNormGal_PAH']
		nuLnuPAH = normPAH * templ['gal_PAH'].values
		enuLnuPAH = normPAH * templ['egal_PAH'].values

		nuLnuGal = nuLnuDust + nuLnuPAH
		enuLnuGal = np.sqrt(enuLnuDust**2. + enuLnuPAH**2.)

		# Calculate the observed model fluxes
		FnuGal = nuLnuToFnu(wavTempl, nuLnuGal, z).values
		eFnuGal = nuLnuToFnu(wavTempl, enuLnuGal, z)

		# Define the optimised model AGN template
		if obj['AGNon'] == 1:
			normAGN = 10**obj['logNormAGN']
			nuLnuAGN = normAGN * templ[obj['tplName_AGN']].values
			
			normSi = 10**obj['logNormSiem']
			nuLnuSi = normSi * templ[nameTempl_Siem].values.flatten()
			
			FnuAGN =  nuLnuToFnu(wavTempl, nuLnuAGN + nuLnuSi, z)

			FnuTot = FnuGal + FnuAGN
		else:
			FnuTot = FnuGal

		# Plot with all the possible models.
		o = np.where(UL < 1.)[0]
		axs1[count].errorbar(wav[o]/(1.+z), flux[o], yerr = eflux[o], fmt = '.', color = '#292F33', alpha = 1.0, label = 'Observed SED')
		o = np.where(UL > 0.)[0]
		axs1[count].errorbar(wav[o]/(1.+z), flux[o], yerr = flux[o]/3., fmt = '.', color = '#292F33', alpha = 1.0, label = '_no_label_', uplims = True)

		if obj['AGNon'] == 1:
			axs1[count].plot(wavTempl, FnuGal, '--', color = '#66757F', label = 'Galaxy comp. ('+obj['tplName_gal']+')')
			axs1[count].plot(wavTempl, FnuAGN, '-.', color = '#66757F', label = 'AGN comp. ('+obj['tplName_AGN']+')')
			if obj['bestModelFlag'] == 1.:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = '#55ACEE', ecolor = 'k', \
								linewidth = 3, elinewidth=0.5, label = 'Total (AGN + Galaxy)', errorevery = 3)
			else:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = 'k', ecolor = 'k', \
									linewidth = 1, elinewidth=0.5, label = 'Total (AGN + Galaxy)', errorevery = 3)
		else:
			if obj['bestModelFlag'] == 1.:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = '#55ACEE', ecolor = 'k', \
								linewidth = 3, elinewidth=0.5, label = 'Total ('+obj['tplName_gal']+')', errorevery = 3)
			else:
				axs1[count].errorbar(wavTempl, FnuTot, yerr = eFnuGal, fmt = '-', color = 'k', ecolor = 'k', \
									linewidth = 1, elinewidth=0.5, label = 'Total ('+obj['tplName_gal']+')', errorevery = 3)

		axs1[count].set_xscale('log')
		axs1[count].set_yscale('log')
		axs1[count].set_xlim([3./(1.+z), 800./(1.+z)])
		if NOAGN != True:
			axs1[count].set_ylim([min(flux)/10., max(flux)*20.])
		else:
			axs1[count].set_ylim([min(flux)/5., max(FnuTot)*10.])
		axs1[count].set_xlabel(r'$\lambda_{\rm rest}\ (\mu {\rm m})$', fontsize = 28)
		if (count == 0) or (count == 6) or (count == 12):
			axs1[count].set_ylabel(r'Flux (Jy)', fontsize = 28)
		axs1[count].legend(frameon = False, loc = 'upper left', fontsize = 15)
		if round(obj['Aw']*100.) > 50.:
			axs1[count].text(10., min(flux)/6., 'Aw = '+str(round(obj['Aw']*100.))+'%', fontsize = 25, c = '#ff4d4d')
		else:
			axs1[count].text(10., min(flux)/6., 'Aw = '+str(round(obj['Aw']*100.))+'%', fontsize = 25, c = 'k')
		axs1[count].tick_params(axis='both', labelcolor='k', labelsize = 25, width = 1, size = 15, which = 'major', direction = 'inout')
		axs1[count].tick_params(axis='both', width = 1, size = 10, which = 'minor', direction = 'inout')

		count += 1

		# Plots with the best model only.
		if obj['Aw'] > 0.05:
			if obj['AGNon'] == 1:
				axs2.plot(wavTempl, FnuTot, '-', color = 'k', \
						 linewidth = 1, label = obj['tplName_gal']+' + '+ obj['tplName_AGN'] + ' ('+str(round(obj['Aw']*100.))+'%)', alpha = obj['Aw'])
			else:
				axs2.plot(wavTempl, FnuTot, '-', color = 'k', \
						 linewidth = 1, label = obj['tplName_gal']+' ('+str(round(obj['Aw']*100.))+'%)', alpha = obj['Aw'])

		try:
			FnuTot_Aw += FnuTot * obj['Aw']
			eFnuTot_Aw += (eFnuGal.values*obj['Aw'])**2.
			FnuGal_Aw += FnuGal * obj['Aw']
			if obj['AGNon'] == 1:
				FnuAGN_Aw += FnuAGN * obj['Aw']
		except:
			FnuTot_Aw = FnuTot * obj['Aw']
			eFnuTot_Aw = (eFnuGal.values * obj['Aw'])**2.
			FnuGal_Aw = FnuGal * obj['Aw']
			if obj['AGNon'] == 1:
				FnuAGN_Aw = FnuAGN * obj['Aw']
			else:
				FnuAGN_Aw = FnuGal_Aw * 0.

	axs2.errorbar(wavTempl, FnuTot_Aw, yerr = np.sqrt(eFnuTot_Aw), fmt = '-', color = 'k', elinewidth = 0.5, linewidth = 2, label = 'Best weighted fit [Total]',\
				 ecolor = 'k', alpha = 0.4, errorevery = 3)
	axs2.plot(wavTempl, FnuGal_Aw, '--', color = '#E94B3C', \
						 linewidth = 2, label = 'Best weighted fit [Galaxy]', alpha = 0.8)
	axs2.plot(wavTempl, FnuAGN_Aw, '-.', color = '#6395F2', \
						 linewidth = 2, label = 'Best weighted fit [AGN]', alpha = 0.8)

	o = np.where(UL < 1.)[0]
	axs2.errorbar(wav[o]/(1.+z), flux[o], yerr = eflux[o], fmt = 'o', color = '#292F33', alpha = 0.9, label = 'Observed SED', mfc = 'none', mew = 1, ms = 10)

	o = np.where(UL > 0.)[0]
	axs2.errorbar(wav[o]/(1.+z), flux[o], yerr = flux[o]/5., fmt = 'o', color = '#292F33', alpha = 0.9, label = '_no_label_', uplims = True, mfc = 'none', mew = 1, ms = 10)

	axs2.set_xscale('log')
	axs2.set_yscale('log')
	axs2.set_xlim([3./(1.+z), 800./(1.+z)])
	if NOAGN != True:
		axs2.set_ylim([min(flux)/5., max(flux)*10.])
	else:
		axs2.set_ylim([min(flux)/5., max(FnuGal_Aw)*10.])
	axs2.set_xlabel(r'$\lambda_{\rm rest}\ (\mu {\rm m})$', FontProperties = font0)
	axs2.set_ylabel(r'Flux (Jy)', FontProperties = font0)
	axs2.legend(frameon = False, fontsize = 16, ncol = 2)

	# if saveRes is True, save the figures at the locations pathFig.
	if saveRes == True:
		fig1.savefig(pathFig+sourceName+'_fitResAll_photo.pdf')
		fig2.savefig(pathFig+sourceName+'_fitResBM_photo.pdf')
	else:
		plt.show()
	plt.close('all')


#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="iragnsep", # Replace with your own username
    version="6.1.0",
    author="Emmanuel Bernhard",
    author_email="manu.p.bernhard@gmail.com",
    description="Fits of IR SEDs including AGN contribution.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/ebernhard/iragnsep/src/master/",
    install_requires=['numpy', 'matplotlib', 'astropy', 'scipy', 'pandas', 'emcee', 'numba', 'tqdm'],
    packages=['iragnsep'],
    package_data={'iragnsep': ['Filters/*.csv', 'iragnsep_templ.csv']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

#install_requires=['emcee'],


import os
import iragnsep
import pandas as pd

from .func import *

class modelToSED:
	"""
    This class calculates the synthetic fluxes within a filter given a model nuLnu and a redshift.
    """
	_z = 0.1

	def __init__(self, lambda_mic, nuLnu, z = _z):
		"""Initialise to z = 0.1, if not provided."""
	
		self.lambda_mic = lambda_mic
		self.z = z
		self.nuLnu = nuLnu

	def IRAC1(self):
		filename = 'IRAC1Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency

		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAC2(self):
		filename = 'IRAC2Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency

		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAC3(self):
		filename = 'IRAC3Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency

		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAC4(self):
		filename = 'IRAC4Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency

		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAS12(self):
		filename = 'IRAS12Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def MIPS24(self):
		filename = 'MIPS24Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAS60(self):
		filename = 'IRAS60Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def MIPS70(self):
		filename = 'MIPS70Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def MIPS160(self):
		filename = 'MIPS160Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def PACS70(self):
		filename = 'PACS70Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def PACS100(self):
		filename = 'PACS100Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def IRAS100(self):
		filename = 'IRAS100Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def PACS160(self):
		filename = 'PACS160Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def SPIRE250ps(self):
		filename = 'SPIRE250psFilter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def SPIRE350ps(self):
		filename = 'SPIRE350psFilter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def SPIRE500ps(self):
		filename = 'SPIRE500psFilter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def WISE_W1(self):
		filename = 'WISE_W1Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def WISE_W2(self):
		filename = 'WISE_W2Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy


	def WISE_W3(self):
		filename = 'WISE_W3Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy

	def WISE_W4(self):
		filename = 'WISE_W4Filter.csv'

		path = os.path.dirname(iragnsep.__file__)+'/Filters/'
		filt = pd.read_csv(path+filename)  # (Energy Counter)
		filt_wav = filt.lambda_ang*1e-4 #Micron (Window of the filter)
		filt_QE = filt.QE # Quantum Efficiency
		
		flux = getFluxInFilt(filt_wav, filt_QE, self.lambda_mic, self.nuLnu, self.z)

		return flux #Jy







